import numpy as np

from abmarl.sim.modules import GridResources


def test_builder():
    sim = GridResources.build()
    assert sim.region == 10
    assert sim.max_value == 1.
    assert sim.min_value == 0.1
    assert sim.revive_rate == 0.04
    assert sim.coverage == 0.75


def test_builder_custom():
    sim = GridResources.build({
        'region': 5,
        'max_value': 2.,
        'min_value': 0.01,
        'revive_rate': 0.5,
        'coverage': 0.4
    })
    assert sim.region == 5
    assert sim.max_value == 2.
    assert sim.min_value == 0.01
    assert sim.revive_rate == 0.5
    assert sim.coverage == 0.4


def test_reset():
    np.random.seed(24)
    sim = GridResources.build({'region': 5})
    sim.reset()
    assert ((sim.resources <= sim.max_value) & (sim.resources >= 0.)).all()


def test_harvest_and_regrow():
    np.random.seed(24)
    sim = GridResources.build()
    sim.reset()

    # Normal action with harvest and replenish
    value_before = {
        (4,5) : sim.resources[(4,5)],
        (3,3) : sim.resources[(3,3)]
    }
    assert sim.harvest((4,5), 0.7) == 0.7
    assert sim.harvest((3,3), 0.1) == 0.1
    sim.regrow()
    assert sim.resources[(4,5)] == value_before[(4,5)] - 0.7 + 0.04
    assert sim.resources[(3,3)] == value_before[(3,3)] - 0.1 + 0.04

    # action that has depleted one of the resources
    value_before = {
        (4,5) : sim.resources[(4,5)],
        (2,1) : sim.resources[(2,1)]
    }
    assert sim.harvest((4,5), 0.7) == value_before[(4,5)]
    assert sim.harvest((2,1), 0.15) == 0.15
    sim.regrow()
    assert sim.resources[(4,5)] == 0.
    assert sim.resources[(2,1)] == value_before[(2,1)] - 0.15 + 0.04

    # Check that the depleted resources do not restore
    value_before = {
        (2,1) : sim.resources[(2,1)]
    }
    sim.regrow()
    assert sim.resources[(4,5)] == 0.
    assert sim.resources[(2,1)] == value_before[(2,1)] + 0.04

    # Check that nothing is above maximum value
    for _ in range(25):
        sim.regrow()
    assert (sim.resources <= sim.max_value).all()



import numpy as np

from abmarl.sim.predator_prey import PredatorPreySimulation, Predator, Prey
from abmarl.managers import TurnBasedManager


def test_turn_based_predator_prey_distance():
    np.random.seed(24)
    predators = [Predator(id=f'predator{i}', attack=1) for i in range(2)]
    prey = [Prey(id=f'prey{i}') for i in range(7)]
    agents = predators + prey
    sim_config = {
        'region': 6,
        'observation_mode': PredatorPreySimulation.ObservationMode.DISTANCE,
        'agents': agents,
    }
    sim = PredatorPreySimulation.build(sim_config)
    sim = TurnBasedManager(sim)

    # Little hackish here because I have to explicitly set their values
    obs = sim.reset()
    sim.agents['predator0'].position = np.array([2, 3])
    sim.agents['predator1'].position = np.array([0, 1])
    sim.agents['prey0'].position = np.array([1, 1])
    sim.agents['prey1'].position = np.array([4, 3])
    sim.agents['prey2'].position = np.array([4, 3])
    sim.agents['prey3'].position = np.array([2, 3])
    sim.agents['prey4'].position = np.array([3, 3])
    sim.agents['prey5'].position = np.array([3, 1])
    sim.agents['prey6'].position = np.array([2, 1])
    obs = {'predator0': sim.sim.get_obs('predator0')}

    np.testing.assert_array_equal(obs['predator0']['predator1'], np.array([-2, -2,  2]))
    np.testing.assert_array_equal(obs['predator0']['prey0'], np.array([-1, -2,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey1'], np.array([2, 0,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey2'], np.array([2, 0,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey3'], np.array([0, 0,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey4'], np.array([1, 0,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey5'], np.array([1, -2,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey6'], np.array([0, -2,  1]))

    obs, reward, done, info = sim.step(
        {agent_id: {'attack': 1, 'move': np.array([0, 0])} for agent_id in obs}
    )
    np.testing.assert_array_equal(obs['predator1']['predator0'], np.array([2, 2,  2]))
    np.testing.assert_array_equal(obs['predator1']['prey0'], np.array([1, 0,  1]))
    np.testing.assert_array_equal(obs['predator1']['prey1'], np.array([4, 2,  1]))
    np.testing.assert_array_equal(obs['predator1']['prey2'], np.array([4, 2,  1]))
    np.testing.assert_array_equal(obs['predator1']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey4'], np.array([3, 2,  1]))
    np.testing.assert_array_equal(obs['predator1']['prey5'], np.array([3, 0,  1]))
    np.testing.assert_array_equal(obs['predator1']['prey6'], np.array([2, 0,  1]))
    assert reward == {'predator1': 0}
    assert done == {'predator1': False, '__all__': False}

    obs, reward, done, info = sim.step(
        {agent_id: {'attack': 1, 'move': np.array([0, 0])} for agent_id in obs}
    )
    np.testing.assert_array_equal(obs['prey0']['predator0'], np.array([1, 2,  2]))
    np.testing.assert_array_equal(obs['prey0']['predator1'], np.array([-1, 0,  2]))
    np.testing.assert_array_equal(obs['prey0']['prey1'], np.array([3, 2,  1]))
    np.testing.assert_array_equal(obs['prey0']['prey2'], np.array([3, 2,  1]))
    np.testing.assert_array_equal(obs['prey0']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey0']['prey4'], np.array([2, 2,  1]))
    np.testing.assert_array_equal(obs['prey0']['prey5'], np.array([2, 0,  1]))
    np.testing.assert_array_equal(obs['prey0']['prey6'], np.array([1, 0,  1]))
    np.testing.assert_array_equal(obs['prey1']['predator0'], np.array([-2, 0,  2]))
    np.testing.assert_array_equal(obs['prey1']['predator1'], np.array([-4, -2,  2]))
    np.testing.assert_array_equal(obs['prey1']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey1']['prey2'], np.array([0, 0,  1]))
    np.testing.assert_array_equal(obs['prey1']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey1']['prey4'], np.array([-1, 0,  1]))
    np.testing.assert_array_equal(obs['prey1']['prey5'], np.array([-1, -2,  1]))
    np.testing.assert_array_equal(obs['prey1']['prey6'], np.array([-2, -2,  1]))
    assert reward == {'prey0': -36, 'prey1': 0}
    assert done == {'prey0': True, 'prey1': False, '__all__': False}

    obs, reward, done, info = sim.step({'prey1': np.array([0, -1])})
    np.testing.assert_array_equal(obs['prey2']['predator0'], np.array([-2, 0,  2]))
    np.testing.assert_array_equal(obs['prey2']['predator1'], np.array([-4, -2,  2]))
    np.testing.assert_array_equal(obs['prey2']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey2']['prey1'], np.array([0, -1,  1]))
    np.testing.assert_array_equal(obs['prey2']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey2']['prey4'], np.array([-1, 0,  1]))
    np.testing.assert_array_equal(obs['prey2']['prey5'], np.array([-1, -2,  1]))
    np.testing.assert_array_equal(obs['prey2']['prey6'], np.array([-2, -2,  1]))
    assert reward == {'prey2': 0}
    assert done == {'prey2': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: np.array([1, 1]) for agent_id in obs})
    np.testing.assert_array_equal(obs['prey3']['predator0'], np.array([0, 0,  2]))
    np.testing.assert_array_equal(obs['prey3']['predator1'], np.array([-2, -2,  2]))
    np.testing.assert_array_equal(obs['prey3']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey3']['prey1'], np.array([2, -1,  1]))
    np.testing.assert_array_equal(obs['prey3']['prey2'], np.array([3, 1,  1]))
    np.testing.assert_array_equal(obs['prey3']['prey4'], np.array([1, 0,  1]))
    np.testing.assert_array_equal(obs['prey3']['prey5'], np.array([1, -2,  1]))
    np.testing.assert_array_equal(obs['prey3']['prey6'], np.array([0, -2,  1]))
    np.testing.assert_array_equal(obs['prey4']['predator0'], np.array([-1, 0,  2]))
    np.testing.assert_array_equal(obs['prey4']['predator1'], np.array([-3, -2,  2]))
    np.testing.assert_array_equal(obs['prey4']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey4']['prey1'], np.array([1, -1,  1]))
    np.testing.assert_array_equal(obs['prey4']['prey2'], np.array([2, 1,  1]))
    np.testing.assert_array_equal(obs['prey4']['prey3'], np.array([-0, 0,  0]))
    np.testing.assert_array_equal(obs['prey4']['prey5'], np.array([0, -2,  1]))
    np.testing.assert_array_equal(obs['prey4']['prey6'], np.array([-1, -2,  1]))
    assert reward == {'prey3': -36, 'prey4': 0}
    assert done == {'prey3': True, 'prey4': False, '__all__': False}

    obs, reward, done, info = sim.step({'prey4': np.array([-1, 1])})
    np.testing.assert_array_equal(obs['prey5']['predator0'], np.array([-1, 2,  2]))
    np.testing.assert_array_equal(obs['prey5']['predator1'], np.array([-3, 0,  2]))
    np.testing.assert_array_equal(obs['prey5']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey5']['prey1'], np.array([1, 1,  1]))
    np.testing.assert_array_equal(obs['prey5']['prey2'], np.array([2, 3,  1]))
    np.testing.assert_array_equal(obs['prey5']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey5']['prey4'], np.array([-1, 3,  1]))
    np.testing.assert_array_equal(obs['prey5']['prey6'], np.array([-1, 0,  1]))
    assert reward == {'prey5': 0}
    assert done == {'prey5': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: np.array([1, 1]) for agent_id in obs})
    np.testing.assert_array_equal(obs['prey6']['predator0'], np.array([0, 2,  2]))
    np.testing.assert_array_equal(obs['prey6']['predator1'], np.array([-2, 0,  2]))
    np.testing.assert_array_equal(obs['prey6']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey6']['prey1'], np.array([2, 1,  1]))
    np.testing.assert_array_equal(obs['prey6']['prey2'], np.array([3, 3,  1]))
    np.testing.assert_array_equal(obs['prey6']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey6']['prey4'], np.array([0, 3,  1]))
    np.testing.assert_array_equal(obs['prey6']['prey5'], np.array([2, 1,  1]))
    assert reward == {'prey6': 0}
    assert done == {'prey6': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: np.array([0, 0]) for agent_id in obs})
    np.testing.assert_array_equal(obs['predator0']['predator1'], np.array([-2, -2,  2]))
    np.testing.assert_array_equal(obs['predator0']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey1'], np.array([2, -1,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey2'], np.array([3, 1,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey4'], np.array([0, 1,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey5'], np.array([2, -1,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey6'], np.array([0, -2,  1]))
    assert reward == {'predator0':36}
    assert done == {'predator0': False, '__all__': False}

    obs, reward, done, info = sim.step(
        {agent_id: {'attack': 1, 'move': np.array([0, 0])} for agent_id in obs}
    )
    np.testing.assert_array_equal(obs['predator1']['predator0'], np.array([2, 2,  2]))
    np.testing.assert_array_equal(obs['predator1']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey1'], np.array([4, 1,  1]))
    np.testing.assert_array_equal(obs['predator1']['prey2'], np.array([5, 3,  1]))
    np.testing.assert_array_equal(obs['predator1']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey4'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey5'], np.array([4, 1,  1]))
    np.testing.assert_array_equal(obs['predator1']['prey6'], np.array([2, 0,  1]))
    assert reward == {'predator1': 36}
    assert done == {'predator1': False, '__all__': False}

    obs, reward, done, info = sim.step(
        {agent_id: {'attack': 0, 'move': np.array([1, 0])} for agent_id in obs}
    )
    np.testing.assert_array_equal(obs['prey1']['predator0'], np.array([-2, 1,  2]))
    np.testing.assert_array_equal(obs['prey1']['predator1'], np.array([-3, -1,  2]))
    np.testing.assert_array_equal(obs['prey1']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey1']['prey2'], np.array([1, 2,  1]))
    np.testing.assert_array_equal(obs['prey1']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey1']['prey4'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey1']['prey5'], np.array([0, 0,  1]))
    np.testing.assert_array_equal(obs['prey1']['prey6'], np.array([-2, -1,  1]))
    assert reward == {'prey1': -1}
    assert done == {'prey1': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: np.array([-1, -1]) for agent_id in obs})
    np.testing.assert_array_equal(obs['prey2']['predator0'], np.array([-3, -1,  2]))
    np.testing.assert_array_equal(obs['prey2']['predator1'], np.array([-4, -3,  2]))
    np.testing.assert_array_equal(obs['prey2']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey2']['prey1'], np.array([-2, -3,  1]))
    np.testing.assert_array_equal(obs['prey2']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey2']['prey4'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey2']['prey5'], np.array([-1, -2,  1]))
    np.testing.assert_array_equal(obs['prey2']['prey6'], np.array([-3, -3,  1]))
    assert reward == {'prey2': -1}
    assert done == {'prey2': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: np.array([-1, 0]) for agent_id in obs})
    np.testing.assert_array_equal(obs['prey4']['predator0'], np.array([0, -1,  2]))
    np.testing.assert_array_equal(obs['prey4']['predator1'], np.array([-1, -3,  2]))
    np.testing.assert_array_equal(obs['prey4']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey4']['prey1'], np.array([1, -3,  1]))
    np.testing.assert_array_equal(obs['prey4']['prey2'], np.array([2, 0,  1]))
    np.testing.assert_array_equal(obs['prey4']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey4']['prey5'], np.array([2, -2,  1]))
    np.testing.assert_array_equal(obs['prey4']['prey6'], np.array([0, -3,  1]))
    np.testing.assert_array_equal(obs['prey5']['predator0'], np.array([-2, 1,  2]))
    np.testing.assert_array_equal(obs['prey5']['predator1'], np.array([-3, -1,  2]))
    np.testing.assert_array_equal(obs['prey5']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey5']['prey1'], np.array([-1, -1,  1]))
    np.testing.assert_array_equal(obs['prey5']['prey2'], np.array([0, 2,  1]))
    np.testing.assert_array_equal(obs['prey5']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey5']['prey4'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(obs['prey5']['prey6'], np.array([-2, -1,  1]))
    assert reward == {'prey4': -37, 'prey5': -1}
    assert done == {'prey4': True, 'prey5': False, '__all__': False}

    obs, reward, done, info = sim.step({'prey5': np.array([-1, 0])})
    np.testing.assert_array_equal(obs['prey6']['predator0'], np.array([0, 2,  2]))
    np.testing.assert_array_equal(obs['prey6']['predator1'], np.array([-1, 0,  2]))
    np.testing.assert_array_equal(obs['prey6']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey6']['prey1'], np.array([1, 0,  1]))
    np.testing.assert_array_equal(obs['prey6']['prey2'], np.array([2, 3,  1]))
    np.testing.assert_array_equal(obs['prey6']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey6']['prey4'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey6']['prey5'], np.array([1, 1,  1]))
    assert reward == {'prey6': 0}
    assert done == {'prey6': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: np.array([0, -1]) for agent_id in obs})
    np.testing.assert_array_equal(obs['predator0']['predator1'], np.array([-1, -2,  2]))
    np.testing.assert_array_equal(obs['predator0']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey1'], np.array([1, -2,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey2'], np.array([2, 1,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey4'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey5'], np.array([1, -1,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey6'], np.array([0, -3,  1]))
    assert reward == {'predator0': 36}
    assert done == {'predator0': False, '__all__': False}

    obs, reward, done, info = sim.step(
        {agent_id: {'attack': 1, 'move': np.array([0, 0])} for agent_id in obs}
    )
    np.testing.assert_array_equal(obs['predator1']['predator0'], np.array([1, 2,  2]))
    np.testing.assert_array_equal(obs['predator1']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey1'], np.array([2, 0,  1]))
    np.testing.assert_array_equal(obs['predator1']['prey2'], np.array([3, 3,  1]))
    np.testing.assert_array_equal(obs['predator1']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey4'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey5'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey6'], np.array([1, -1,  1]))
    assert reward == {'predator1': -1}
    assert done == {'predator1': False, '__all__': False}

    obs, reward, done, info = sim.step(
        {agent_id: {'attack': 1, 'move': np.array([0, 0])} for agent_id in obs}
    )
    np.testing.assert_array_equal(obs['prey1']['predator0'], np.array([-1, 2,  2]))
    np.testing.assert_array_equal(obs['prey1']['predator1'], np.array([-2, 0,  2]))
    np.testing.assert_array_equal(obs['prey1']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey1']['prey2'], np.array([1, 3,  1]))
    np.testing.assert_array_equal(obs['prey1']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey1']['prey4'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey1']['prey5'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey1']['prey6'], np.array([0, 0,  0]))
    assert reward == {'prey1': -1}
    assert done == {'prey1': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: np.array([-1, 0]) for agent_id in obs})
    np.testing.assert_array_equal(obs['prey2']['predator0'], np.array([-2, -1,  2]))
    np.testing.assert_array_equal(obs['prey2']['predator1'], np.array([-3, -3,  2]))
    np.testing.assert_array_equal(obs['prey2']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey2']['prey1'], np.array([-2, -3,  1]))
    np.testing.assert_array_equal(obs['prey2']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey2']['prey4'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey2']['prey5'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey2']['prey6'], np.array([0, 0,  0]))
    assert reward == {'prey2': -1}
    assert done == {'prey2': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: np.array([-1, 0]) for agent_id in obs})
    np.testing.assert_array_equal(obs['prey5']['predator0'], np.array([-1, 1,  2]))
    np.testing.assert_array_equal(obs['prey5']['predator1'], np.array([-2, -1,  2]))
    np.testing.assert_array_equal(obs['prey5']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey5']['prey1'], np.array([-1, -1,  1]))
    np.testing.assert_array_equal(obs['prey5']['prey2'], np.array([0, 2,  1]))
    np.testing.assert_array_equal(obs['prey5']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey5']['prey4'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(obs['prey5']['prey6'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey6']['predator0'], np.array([0, 3,  2]))
    np.testing.assert_array_equal(obs['prey6']['predator1'], np.array([-1, 1,  2]))
    np.testing.assert_array_equal(obs['prey6']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey6']['prey1'], np.array([0, 1,  1]))
    np.testing.assert_array_equal(obs['prey6']['prey2'], np.array([1, 4,  1]))
    np.testing.assert_array_equal(obs['prey6']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey6']['prey4'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey6']['prey5'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['predator1'], np.array([-1, -2,  2]))
    np.testing.assert_array_equal(obs['predator0']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey1'], np.array([0, -2,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey2'], np.array([1, 1,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey4'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey5'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey6'], np.array([0, 0,  0]))
    assert reward == {'prey5': -37, 'prey6': -37, 'predator0': 36}
    assert done == {'prey5': True, 'prey6': True, 'predator0': False, '__all__': False}

    obs, reward, done, info = sim.step({'predator0': {'attack': 1, 'move': np.array([0, 0])}})
    np.testing.assert_array_equal(obs['predator1']['predator0'], np.array([1, 2,  2]))
    np.testing.assert_array_equal(obs['predator1']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey1'], np.array([1, 0,  1]))
    np.testing.assert_array_equal(obs['predator1']['prey2'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey4'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey5'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey6'], np.array([0, 0,  0]))
    assert reward == {'predator1': 36}
    assert done == {'predator1': False, '__all__': False}

    obs, reward, done, info = sim.step(
        {agent_id: {'attack': 1, 'move': np.array([0, 0])} for agent_id in obs}
    )
    np.testing.assert_array_equal(obs['prey1']['predator0'], np.array([0, 2,  2]))
    np.testing.assert_array_equal(obs['prey1']['predator1'], np.array([-1, 0,  2]))
    np.testing.assert_array_equal(obs['prey1']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey1']['prey2'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(obs['prey1']['prey3'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(obs['prey1']['prey4'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(obs['prey1']['prey5'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(obs['prey1']['prey6'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(obs['prey2']['predator0'], np.array([-1, -1,  2]))
    np.testing.assert_array_equal(obs['prey2']['predator1'], np.array([-2, -3,  2]))
    np.testing.assert_array_equal(obs['prey2']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey2']['prey1'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey2']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey2']['prey4'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey2']['prey5'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey2']['prey6'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['predator1'], np.array([-1, -2,  2]))
    np.testing.assert_array_equal(obs['predator0']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey1'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey2'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey4'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey5'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey6'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['predator0'], np.array([1, 2,  2]))
    np.testing.assert_array_equal(obs['predator1']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey1'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey2'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey4'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey5'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey6'], np.array([0, 0,  0]))
    assert reward == {'prey1': -37, 'prey2': -37, 'predator0': 36, 'predator1': 36}
    assert done == {
        'prey1': True, 'prey2': True, 'predator0': False, 'predator1': False, '__all__': True
    }


def test_turn_based_predator_prey_grid():
    np.random.seed(24)
    predators = [Predator(id=f'predator{i}', attack=1, view=0) for i in range(2)]
    prey = [Prey(id=f'prey{i}', view=0) for i in range(7)]
    agents = predators + prey
    sim_config = {
        'region': 6,
        'observation_mode': PredatorPreySimulation.ObservationMode.GRID,
        'agents': agents,
    }
    sim = PredatorPreySimulation.build(sim_config)
    sim = TurnBasedManager(sim)

    # Little hackish here because I have to explicitly set their values
    obs = sim.reset()
    sim.agents['predator0'].position = np.array([2, 3])
    sim.agents['predator1'].position = np.array([0, 1])
    sim.agents['prey0'].position = np.array([1, 1])
    sim.agents['prey1'].position = np.array([4, 3])
    sim.agents['prey2'].position = np.array([4, 3])
    sim.agents['prey3'].position = np.array([2, 3])
    sim.agents['prey4'].position = np.array([3, 3])
    sim.agents['prey5'].position = np.array([3, 1])
    sim.agents['prey6'].position = np.array([2, 1])
    obs = {'predator0': sim.sim.get_obs('predator0')}

    assert len(obs) == 1 and 'predator0' in obs

    obs, reward, done, info = sim.step(
        {agent_id: {'attack': 1, 'move': np.array([0, 0])} for agent_id in obs}
    )
    assert len(obs) == 1 and 'predator1' in obs
    assert reward == {'predator1': 0}
    assert done == {'predator1': False, '__all__': False}

    obs, reward, done, info = sim.step(
        {agent_id: {'attack': 1, 'move': np.array([0, 0])} for agent_id in obs}
    )
    assert len(obs) == 2 and 'prey0' in obs and 'prey1' in obs
    assert reward == {'prey0': -36, 'prey1': 0}
    assert done == {'prey0': True, 'prey1': False, '__all__': False}

    obs, reward, done, info = sim.step(
        {'prey1': {'move': np.array([0, -1]), 'harvest': 0}}
    )
    assert len(obs) == 1 and 'prey2' in obs
    assert reward == {'prey2': 0}
    assert done == {'prey2': False, '__all__': False}

    obs, reward, done, info = sim.step(
        {agent_id: {'move': np.array([1, 1]), 'harvest': 0} for agent_id in obs}
    )
    assert len(obs) == 2 and 'prey3' in obs and 'prey4' in obs
    assert reward == {'prey3': -36, 'prey4': 0}
    assert done == {'prey3': True, 'prey4': False, '__all__': False}

    obs, reward, done, info = sim.step(
        {'prey4': {'move': np.array([-1, 1]), 'harvest': 0} for agent_id in obs}
    )
    assert len(obs) == 1 and 'prey5' in obs
    assert reward == {'prey5': 0}
    assert done == {'prey5': False, '__all__': False}

    obs, reward, done, info = sim.step(
        {agent_id: {'move': np.array([1, 1]), 'harvest': 0} for agent_id in obs}
    )
    assert len(obs) == 1 and 'prey6' in obs
    assert reward == {'prey6': 0}
    assert done == {'prey6': False, '__all__': False}

    obs, reward, done, info = sim.step(
        {agent_id: {'move': np.array([0, 0]), 'harvest': 0} for agent_id in obs}
    )
    assert len(obs) == 1 and 'predator0' in obs
    assert reward == {'predator0':36}
    assert done == {'predator0': False, '__all__': False}

    obs, reward, done, info = sim.step(
        {agent_id: {'attack': 1, 'move': np.array([0, 0])} for agent_id in obs}
    )
    assert len(obs) == 1 and 'predator1' in obs
    assert reward == {'predator1': 36}
    assert done == {'predator1': False, '__all__': False}

    obs, reward, done, info = sim.step(
        {agent_id: {'attack': 0, 'move': np.array([1, 0])} for agent_id in obs}
    )
    assert len(obs) == 1 and 'prey1' in obs
    assert reward == {'prey1': -1}
    assert done == {'prey1': False, '__all__': False}

    obs, reward, done, info = sim.step(
        {agent_id: {'move': np.array([-1, -1]), 'harvest': 0} for agent_id in obs}
    )
    assert len(obs) == 1 and 'prey2' in obs
    assert reward == {'prey2': -1}
    assert done == {'prey2': False, '__all__': False}

    obs, reward, done, info = sim.step(
        {agent_id: {'move': np.array([-1, 0]), 'harvest': 0} for agent_id in obs}
    )
    assert len(obs) == 2 and 'prey4' in obs and 'prey5'
    assert reward == {'prey4': -37, 'prey5': -1}
    assert done == {'prey4': True, 'prey5': False, '__all__': False}

    obs, reward, done, info = sim.step(
        {'prey5': {'move': np.array([-1, 0]), 'harvest': 0} for agent_id in obs}
    )
    assert len(obs) == 1 and 'prey6' in obs
    assert reward == {'prey6': 0}
    assert done == {'prey6': False, '__all__': False}

    obs, reward, done, info = sim.step(
        {agent_id: {'move': np.array([0, -1]), 'harvest': 0} for agent_id in obs}
    )
    assert len(obs) == 1 and 'predator0' in obs
    assert reward == {'predator0': 36}
    assert done == {'predator0': False, '__all__': False}

    obs, reward, done, info = sim.step(
        {agent_id: {'attack': 1, 'move': np.array([0, 0])} for agent_id in obs}
    )
    assert len(obs) == 1 and 'predator1' in obs
    assert reward == {'predator1': -1}
    assert done == {'predator1': False, '__all__': False}

    obs, reward, done, info = sim.step(
        {agent_id: {'attack': 1, 'move': np.array([0, 0])} for agent_id in obs}
    )
    assert len(obs) == 1 and 'prey1' in obs
    assert reward == {'prey1': -1}
    assert done == {'prey1': False, '__all__': False}

    obs, reward, done, info = sim.step(
        {agent_id: {'move': np.array([-1, 0]), 'harvest': 0} for agent_id in obs}
    )
    assert len(obs) == 1 and 'prey2' in obs
    assert reward == {'prey2': -1}
    assert done == {'prey2': False, '__all__': False}

    obs, reward, done, info = sim.step(
        {agent_id: {'move': np.array([-1, 0]), 'harvest': 0} for agent_id in obs}
    )
    assert len(obs) == 3 and 'prey5' in obs and 'prey6' in obs and 'predator0' in obs
    assert reward == {'prey5': -37, 'prey6': -37, 'predator0': 36}
    assert done == {'prey5': True, 'prey6': True, 'predator0': False, '__all__': False}

    obs, reward, done, info = sim.step({'predator0': {'attack': 1, 'move': np.array([0, 0])}})
    assert len(obs) == 1 and 'predator1' in obs
    assert reward == {'predator1': 36}
    assert done == {'predator1': False, '__all__': False}

    obs, reward, done, info = sim.step(
        {agent_id: {'attack': 1, 'move': np.array([0, 0])} for agent_id in obs}
    )
    assert len(obs) == 4
    assert 'prey1' in obs
    assert 'prey2' in obs
    assert 'predator0' in obs
    assert 'predator1' in obs
    assert reward == {'prey1': -37, 'prey2': -37, 'predator0': 36, 'predator1': 36}
    assert done == {
        'prey1': True, 'prey2': True, 'predator0': False, 'predator1': False, '__all__': True
    }



import numpy as np
import pytest

from abmarl.sim.predator_prey import PredatorPreySimulation, Predator, Prey
from abmarl.managers import AllStepManager


def test_turn_based_predator_prey_distance():
    np.random.seed(24)
    predators = [Predator(id=f'predator{i}', attack=1) for i in range(2)]
    prey = [Prey(id=f'prey{i}') for i in range(7)]
    agents = predators + prey
    sim_config = {
        'region': 6,
        'observation_mode': PredatorPreySimulation.ObservationMode.DISTANCE,
        'agents': agents,
    }
    sim = PredatorPreySimulation.build(sim_config)
    sim = AllStepManager(sim)

    # Little hackish here because I have to explicitly set their values
    obs = sim.reset()
    sim.agents['predator0'].position = np.array([2, 3])
    sim.agents['predator1'].position = np.array([0, 1])
    sim.agents['prey0'].position = np.array([1, 1])
    sim.agents['prey1'].position = np.array([4, 3])
    sim.agents['prey2'].position = np.array([4, 3])
    sim.agents['prey3'].position = np.array([2, 3])
    sim.agents['prey4'].position = np.array([3, 3])
    sim.agents['prey5'].position = np.array([3, 1])
    sim.agents['prey6'].position = np.array([2, 1])
    obs = {agent_id: sim.sim.get_obs(agent_id) for agent_id in sim.agents}

    np.testing.assert_array_equal(obs['predator0']['predator1'], np.array([-2, -2,  2]))
    np.testing.assert_array_equal(obs['predator0']['prey0'], np.array([-1, -2,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey1'], np.array([2, 0,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey2'], np.array([2, 0,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey3'], np.array([0, 0,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey4'], np.array([1, 0,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey5'], np.array([1, -2,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey6'], np.array([0, -2,  1]))

    np.testing.assert_array_equal(obs['predator1']['predator0'], np.array([2, 2,  2]))
    np.testing.assert_array_equal(obs['predator1']['prey0'], np.array([1, 0,  1]))
    np.testing.assert_array_equal(obs['predator1']['prey1'], np.array([4, 2,  1]))
    np.testing.assert_array_equal(obs['predator1']['prey2'], np.array([4, 2,  1]))
    np.testing.assert_array_equal(obs['predator1']['prey3'], np.array([2, 2,  1]))
    np.testing.assert_array_equal(obs['predator1']['prey4'], np.array([3, 2,  1]))
    np.testing.assert_array_equal(obs['predator1']['prey5'], np.array([3, 0,  1]))
    np.testing.assert_array_equal(obs['predator1']['prey6'], np.array([2, 0,  1]))

    np.testing.assert_array_equal(obs['prey0']['predator0'], np.array([1, 2,  2]))
    np.testing.assert_array_equal(obs['prey0']['predator1'], np.array([-1, 0,  2]))
    np.testing.assert_array_equal(obs['prey0']['prey1'], np.array([3, 2,  1]))
    np.testing.assert_array_equal(obs['prey0']['prey2'], np.array([3, 2,  1]))
    np.testing.assert_array_equal(obs['prey0']['prey3'], np.array([1, 2, 1]))
    np.testing.assert_array_equal(obs['prey0']['prey4'], np.array([2, 2,  1]))
    np.testing.assert_array_equal(obs['prey0']['prey5'], np.array([2, 0,  1]))
    np.testing.assert_array_equal(obs['prey0']['prey6'], np.array([1, 0,  1]))

    np.testing.assert_array_equal(obs['prey1']['predator0'], np.array([-2, 0,  2]))
    np.testing.assert_array_equal(obs['prey1']['predator1'], np.array([-4, -2,  2]))
    np.testing.assert_array_equal(obs['prey1']['prey0'], np.array([-3, -2,  1]))
    np.testing.assert_array_equal(obs['prey1']['prey2'], np.array([0, 0,  1]))
    np.testing.assert_array_equal(obs['prey1']['prey3'], np.array([-2, 0, 1]))
    np.testing.assert_array_equal(obs['prey1']['prey4'], np.array([-1, 0,  1]))
    np.testing.assert_array_equal(obs['prey1']['prey5'], np.array([-1, -2,  1]))
    np.testing.assert_array_equal(obs['prey1']['prey6'], np.array([-2, -2,  1]))

    np.testing.assert_array_equal(obs['prey2']['predator0'], np.array([-2, 0,  2]))
    np.testing.assert_array_equal(obs['prey2']['predator1'], np.array([-4, -2,  2]))
    np.testing.assert_array_equal(obs['prey2']['prey0'], np.array([-3, -2, 1]))
    np.testing.assert_array_equal(obs['prey2']['prey1'], np.array([0, 0,  1]))
    np.testing.assert_array_equal(obs['prey2']['prey3'], np.array([-2, 0, 1]))
    np.testing.assert_array_equal(obs['prey2']['prey4'], np.array([-1, 0,  1]))
    np.testing.assert_array_equal(obs['prey2']['prey5'], np.array([-1, -2,  1]))
    np.testing.assert_array_equal(obs['prey2']['prey6'], np.array([-2, -2,  1]))

    np.testing.assert_array_equal(obs['prey3']['predator0'], np.array([0, 0,  2]))
    np.testing.assert_array_equal(obs['prey3']['predator1'], np.array([-2, -2,  2]))
    np.testing.assert_array_equal(obs['prey3']['prey0'], np.array([-1, -2, 1]))
    np.testing.assert_array_equal(obs['prey3']['prey1'], np.array([2, 0,  1]))
    np.testing.assert_array_equal(obs['prey3']['prey2'], np.array([2, 0, 1]))
    np.testing.assert_array_equal(obs['prey3']['prey4'], np.array([1, 0,  1]))
    np.testing.assert_array_equal(obs['prey3']['prey5'], np.array([1, -2,  1]))
    np.testing.assert_array_equal(obs['prey3']['prey6'], np.array([0, -2,  1]))

    np.testing.assert_array_equal(obs['prey4']['predator0'], np.array([-1, 0,  2]))
    np.testing.assert_array_equal(obs['prey4']['predator1'], np.array([-3, -2,  2]))
    np.testing.assert_array_equal(obs['prey4']['prey0'], np.array([-2, -2, 1]))
    np.testing.assert_array_equal(obs['prey4']['prey1'], np.array([1, 0,  1]))
    np.testing.assert_array_equal(obs['prey4']['prey2'], np.array([1, 0,  1]))
    np.testing.assert_array_equal(obs['prey4']['prey3'], np.array([-1, 0, 1]))
    np.testing.assert_array_equal(obs['prey4']['prey5'], np.array([0, -2,  1]))
    np.testing.assert_array_equal(obs['prey4']['prey6'], np.array([-1, -2,  1]))

    np.testing.assert_array_equal(obs['prey5']['predator0'], np.array([-1, 2,  2]))
    np.testing.assert_array_equal(obs['prey5']['predator1'], np.array([-3, 0,  2]))
    np.testing.assert_array_equal(obs['prey5']['prey0'], np.array([-2, 0, 1]))
    np.testing.assert_array_equal(obs['prey5']['prey1'], np.array([1, 2,  1]))
    np.testing.assert_array_equal(obs['prey5']['prey2'], np.array([1, 2,  1]))
    np.testing.assert_array_equal(obs['prey5']['prey3'], np.array([-1, 2, 1]))
    np.testing.assert_array_equal(obs['prey5']['prey4'], np.array([0, 2,  1]))
    np.testing.assert_array_equal(obs['prey5']['prey6'], np.array([-1, 0,  1]))

    np.testing.assert_array_equal(obs['prey6']['predator0'], np.array([0, 2,  2]))
    np.testing.assert_array_equal(obs['prey6']['predator1'], np.array([-2, 0,  2]))
    np.testing.assert_array_equal(obs['prey6']['prey0'], np.array([-1, 0, 1]))
    np.testing.assert_array_equal(obs['prey6']['prey1'], np.array([2, 2,  1]))
    np.testing.assert_array_equal(obs['prey6']['prey2'], np.array([2, 2,  1]))
    np.testing.assert_array_equal(obs['prey6']['prey3'], np.array([0, 2, 1]))
    np.testing.assert_array_equal(obs['prey6']['prey4'], np.array([1, 2,  1]))
    np.testing.assert_array_equal(obs['prey6']['prey5'], np.array([1, 0, 1]))


    obs, reward, done, info = sim.step({
        'predator0': {'attack': 1, 'move': np.array([0, 0])},
        'predator1': {'attack': 1, 'move': np.array([0, 0])},
        'prey0': np.array([-1, 1]),
        'prey1': np.array([0, -1]),
        'prey2': np.array([1, 1]),
        'prey3': np.array([1, -1]),
        'prey4': np.array([-1, 1]),
        'prey5': np.array([1, 1]),
        'prey6': np.array([0, 0]),
    })

    np.testing.assert_array_equal(obs['predator0']['predator1'], np.array([-2, -2,  2]))
    np.testing.assert_array_equal(obs['predator0']['prey0'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(obs['predator0']['prey1'], np.array([2, -1,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey2'], np.array([3, 1,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey4'], np.array([0, 1,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey5'], np.array([2, -1,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey6'], np.array([0, -2,  1]))

    np.testing.assert_array_equal(obs['predator1']['predator0'], np.array([2, 2,  2]))
    np.testing.assert_array_equal(obs['predator1']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey1'], np.array([4, 1,  1]))
    np.testing.assert_array_equal(obs['predator1']['prey2'], np.array([5, 3,  1]))
    np.testing.assert_array_equal(obs['predator1']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey4'], np.array([2, 3, 1]))
    np.testing.assert_array_equal(obs['predator1']['prey5'], np.array([4, 1,  1]))
    np.testing.assert_array_equal(obs['predator1']['prey6'], np.array([2, 0,  1]))

    np.testing.assert_array_equal(obs['prey0']['predator0'], np.array([1, 2,  2]))
    np.testing.assert_array_equal(obs['prey0']['predator1'], np.array([-1, 0,  2]))
    np.testing.assert_array_equal(obs['prey0']['prey1'], np.array([3, 1,  1]))
    np.testing.assert_array_equal(obs['prey0']['prey2'], np.array([4, 3,  1]))
    np.testing.assert_array_equal(obs['prey0']['prey3'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(obs['prey0']['prey4'], np.array([1, 3,  1]))
    np.testing.assert_array_equal(obs['prey0']['prey5'], np.array([3, 1,  1]))
    np.testing.assert_array_equal(obs['prey0']['prey6'], np.array([1, 0,  1]))

    np.testing.assert_array_equal(obs['prey1']['predator0'], np.array([-2, 1,  2]))
    np.testing.assert_array_equal(obs['prey1']['predator1'], np.array([-4, -1,  2]))
    np.testing.assert_array_equal(obs['prey1']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey1']['prey2'], np.array([1, 2,  1]))
    np.testing.assert_array_equal(obs['prey1']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey1']['prey4'], np.array([-2, 2, 1]))
    np.testing.assert_array_equal(obs['prey1']['prey5'], np.array([0, 0,  1]))
    np.testing.assert_array_equal(obs['prey1']['prey6'], np.array([-2, -1,  1]))

    np.testing.assert_array_equal(obs['prey2']['predator0'], np.array([-3, -1,  2]))
    np.testing.assert_array_equal(obs['prey2']['predator1'], np.array([-5, -3,  2]))
    np.testing.assert_array_equal(obs['prey2']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey2']['prey1'], np.array([-1, -2,  1]))
    np.testing.assert_array_equal(obs['prey2']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey2']['prey4'], np.array([-3, 0, 1]))
    np.testing.assert_array_equal(obs['prey2']['prey5'], np.array([-1, -2,  1]))
    np.testing.assert_array_equal(obs['prey2']['prey6'], np.array([-3, -3,  1]))

    np.testing.assert_array_equal(obs['prey3']['predator0'], np.array([0, 0, 2]))
    np.testing.assert_array_equal(obs['prey3']['predator1'], np.array([-2, -2,  2]))
    np.testing.assert_array_equal(obs['prey3']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey3']['prey1'], np.array([2, -1,  1]))
    np.testing.assert_array_equal(obs['prey3']['prey2'], np.array([3, 1, 1]))
    np.testing.assert_array_equal(obs['prey3']['prey4'], np.array([0, 1, 1]))
    np.testing.assert_array_equal(obs['prey3']['prey5'], np.array([2, -1,  1]))
    np.testing.assert_array_equal(obs['prey3']['prey6'], np.array([0, -2,  1]))

    np.testing.assert_array_equal(obs['prey4']['predator0'], np.array([0, -1,  2]))
    np.testing.assert_array_equal(obs['prey4']['predator1'], np.array([-2, -3,  2]))
    np.testing.assert_array_equal(obs['prey4']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey4']['prey1'], np.array([2, -2,  1]))
    np.testing.assert_array_equal(obs['prey4']['prey2'], np.array([3, 0,  1]))
    np.testing.assert_array_equal(obs['prey4']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey4']['prey5'], np.array([2, -2,  1]))
    np.testing.assert_array_equal(obs['prey4']['prey6'], np.array([0, -3,  1]))

    np.testing.assert_array_equal(obs['prey5']['predator0'], np.array([-2, 1,  2]))
    np.testing.assert_array_equal(obs['prey5']['predator1'], np.array([-4, -1,  2]))
    np.testing.assert_array_equal(obs['prey5']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey5']['prey1'], np.array([0, 0,  1]))
    np.testing.assert_array_equal(obs['prey5']['prey2'], np.array([1, 2,  1]))
    np.testing.assert_array_equal(obs['prey5']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey5']['prey4'], np.array([-2, 2, 1]))
    np.testing.assert_array_equal(obs['prey5']['prey6'], np.array([-2, -1,  1]))

    np.testing.assert_array_equal(obs['prey6']['predator0'], np.array([0, 2,  2]))
    np.testing.assert_array_equal(obs['prey6']['predator1'], np.array([-2, 0,  2]))
    np.testing.assert_array_equal(obs['prey6']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey6']['prey1'], np.array([2, 1,  1]))
    np.testing.assert_array_equal(obs['prey6']['prey2'], np.array([3, 3,  1]))
    np.testing.assert_array_equal(obs['prey6']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey6']['prey4'], np.array([0, 3, 1]))
    np.testing.assert_array_equal(obs['prey6']['prey5'], np.array([2, 1,  1]))

    assert reward == {
        'predator0': 36,
        'predator1': 36,
        'prey0': -36,
        'prey1': -1,
        'prey2': -1,
        'prey3': -36,
        'prey4': -1,
        'prey5': -1,
        'prey6': 0,
    }

    assert done == {
        'predator0': False,
        'predator1': False,
        'prey0': True,
        'prey1': False,
        'prey2': False,
        'prey3': True,
        'prey4': False,
        'prey5': False,
        'prey6': False,
        '__all__': False}

    with pytest.raises(AssertionError):
        obs, reward, done, info = sim.step({
            'predator0': {'attack': 1, 'move': np.array([0, 0])},
            'predator1': {'attack': 1, 'move': np.array([0, 0])},
            'prey0': np.array([-1, 1]),
            'prey1': np.array([0, -1]),
            'prey2': np.array([1, 1]),
            'prey3': np.array([1, -1]),
            'prey4': np.array([-1, 1]),
            'prey5': np.array([1, 1]),
            'prey6': np.array([0, 0]),
        })

    obs, reward, done, info = sim.step({
        'predator0': {'attack': 1, 'move': np.array([0, 0])},
        'predator1': {'attack': 0, 'move': np.array([1, 0])},
        'prey1': np.array([-1, -1]),
        'prey2': np.array([-1, 0]),
        'prey4': np.array([-1, 0]),
        'prey5': np.array([-1, 0]),
        'prey6': np.array([0, -1]),
    })

    np.testing.assert_array_equal(obs['predator0']['predator1'], np.array([-1, -2,  2]))
    np.testing.assert_array_equal(obs['predator0']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey1'], np.array([1, -2,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey2'], np.array([2, 1,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey4'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey5'], np.array([1, -1,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey6'], np.array([0, -3,  1]))

    np.testing.assert_array_equal(obs['predator1']['predator0'], np.array([1, 2,  2]))
    np.testing.assert_array_equal(obs['predator1']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey1'], np.array([2, 0,  1]))
    np.testing.assert_array_equal(obs['predator1']['prey2'], np.array([3, 3,  1]))
    np.testing.assert_array_equal(obs['predator1']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey4'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey5'], np.array([2, 1, 1]))
    np.testing.assert_array_equal(obs['predator1']['prey6'], np.array([1, -1,  1]))

    np.testing.assert_array_equal(obs['prey1']['predator0'], np.array([-1, 2,  2]))
    np.testing.assert_array_equal(obs['prey1']['predator1'], np.array([-2, 0,  2]))
    np.testing.assert_array_equal(obs['prey1']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey1']['prey2'], np.array([1, 3,  1]))
    np.testing.assert_array_equal(obs['prey1']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey1']['prey4'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey1']['prey5'], np.array([0, 1, 1]))
    np.testing.assert_array_equal(obs['prey1']['prey6'], np.array([-1, -1, 1]))

    np.testing.assert_array_equal(obs['prey2']['predator0'], np.array([-2, -1,  2]))
    np.testing.assert_array_equal(obs['prey2']['predator1'], np.array([-3, -3,  2]))
    np.testing.assert_array_equal(obs['prey2']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey2']['prey1'], np.array([-1, -3,  1]))
    np.testing.assert_array_equal(obs['prey2']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey2']['prey4'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey2']['prey5'], np.array([-1, -2, 1]))
    np.testing.assert_array_equal(obs['prey2']['prey6'], np.array([-2, -4, 1]))

    np.testing.assert_array_equal(obs['prey4']['predator0'], np.array([0, -1,  2]))
    np.testing.assert_array_equal(obs['prey4']['predator1'], np.array([-1, -3,  2]))
    np.testing.assert_array_equal(obs['prey4']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey4']['prey1'], np.array([1, -3,  1]))
    np.testing.assert_array_equal(obs['prey4']['prey2'], np.array([2, 0, 1]))
    np.testing.assert_array_equal(obs['prey4']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey4']['prey5'], np.array([1, -2, 1]))
    np.testing.assert_array_equal(obs['prey4']['prey6'], np.array([0, -4,  1]))

    np.testing.assert_array_equal(obs['prey5']['predator0'], np.array([-1, 1,  2]))
    np.testing.assert_array_equal(obs['prey5']['predator1'], np.array([-2, -1,  2]))
    np.testing.assert_array_equal(obs['prey5']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey5']['prey1'], np.array([0, -1,  1]))
    np.testing.assert_array_equal(obs['prey5']['prey2'], np.array([1, 2,  1]))
    np.testing.assert_array_equal(obs['prey5']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey5']['prey4'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(obs['prey5']['prey6'], np.array([-1, -2, 1]))

    np.testing.assert_array_equal(obs['prey6']['predator0'], np.array([0, 3,  2]))
    np.testing.assert_array_equal(obs['prey6']['predator1'], np.array([-1, 1,  2]))
    np.testing.assert_array_equal(obs['prey6']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey6']['prey1'], np.array([1, 1,  1]))
    np.testing.assert_array_equal(obs['prey6']['prey2'], np.array([2, 4,  1]))
    np.testing.assert_array_equal(obs['prey6']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey6']['prey4'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey6']['prey5'], np.array([1, 2, 1]))

    assert reward == {
        'predator0': 36,
        'predator1': -1,
        'prey1': -1,
        'prey2': -1,
        'prey4': -36,
        'prey5': -1,
        'prey6': -1,
    }

    assert done == {
        'predator0': False,
        'predator1': False,
        'prey1': False,
        'prey2': False,
        'prey4': True,
        'prey5': False,
        'prey6': False,
        '__all__': False}

    with pytest.raises(AssertionError):
        obs, reward, done, info = sim.step({
            'predator0': {'attack': 1, 'move': np.array([0, 0])},
            'predator1': {'attack': 1, 'move': np.array([0, 0])},
            'prey1': np.array([0, -1]),
            'prey2': np.array([1, 1]),
            'prey4': np.array([-1, 1]),
            'prey5': np.array([1, 1]),
            'prey6': np.array([0, 0]),
        })

    obs, reward, done, info = sim.step({
        'predator0': {'attack': 1, 'move': np.array([0, 0])},
        'predator1': {'attack': 1, 'move': np.array([0, 0])},
        'prey1': np.array([-1, 0]),
        'prey2': np.array([-1, 0]),
        'prey5': np.array([0, 1]),
        'prey6': np.array([-1, 0]),
    })

    np.testing.assert_array_equal(obs['predator0']['predator1'], np.array([-1, -2,  2]))
    np.testing.assert_array_equal(obs['predator0']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey1'], np.array([0, -2,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey2'], np.array([1, 1,  1]))
    np.testing.assert_array_equal(obs['predator0']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey4'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey5'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey6'], np.array([0, 0,  0]))

    np.testing.assert_array_equal(obs['predator1']['predator0'], np.array([1, 2,  2]))
    np.testing.assert_array_equal(obs['predator1']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey1'], np.array([1, 0,  1]))
    np.testing.assert_array_equal(obs['predator1']['prey2'], np.array([2, 3, 1]))
    np.testing.assert_array_equal(obs['predator1']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey4'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey5'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey6'], np.array([0, 0,  0]))

    np.testing.assert_array_equal(obs['prey1']['predator0'], np.array([0, 2,  2]))
    np.testing.assert_array_equal(obs['prey1']['predator1'], np.array([-1, 0,  2]))
    np.testing.assert_array_equal(obs['prey1']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey1']['prey2'], np.array([1, 3, 1]))
    np.testing.assert_array_equal(obs['prey1']['prey3'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(obs['prey1']['prey4'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(obs['prey1']['prey5'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(obs['prey1']['prey6'], np.array([0, 0, 0]))

    np.testing.assert_array_equal(obs['prey2']['predator0'], np.array([-1, -1,  2]))
    np.testing.assert_array_equal(obs['prey2']['predator1'], np.array([-2, -3,  2]))
    np.testing.assert_array_equal(obs['prey2']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey2']['prey1'], np.array([-1, -3, 1]))
    np.testing.assert_array_equal(obs['prey2']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey2']['prey4'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey2']['prey5'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey2']['prey6'], np.array([0, 0,  0]))

    np.testing.assert_array_equal(obs['prey5']['predator0'], np.array([-1, 1,  2]))
    np.testing.assert_array_equal(obs['prey5']['predator1'], np.array([-2, -1,  2]))
    np.testing.assert_array_equal(obs['prey5']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey5']['prey1'], np.array([-1, -1,  1]))
    np.testing.assert_array_equal(obs['prey5']['prey2'], np.array([0, 2,  1]))
    np.testing.assert_array_equal(obs['prey5']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey5']['prey4'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(obs['prey5']['prey6'], np.array([0, 0, 0]))

    np.testing.assert_array_equal(obs['prey6']['predator0'], np.array([0, 3,  2]))
    np.testing.assert_array_equal(obs['prey6']['predator1'], np.array([-1, 1,  2]))
    np.testing.assert_array_equal(obs['prey6']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey6']['prey1'], np.array([0, 1,  1]))
    np.testing.assert_array_equal(obs['prey6']['prey2'], np.array([1, 4,  1]))
    np.testing.assert_array_equal(obs['prey6']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey6']['prey4'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey6']['prey5'], np.array([0, 0, 0]))

    assert reward == {
        'predator0': 36,
        'predator1': 36,
        'prey1': -1,
        'prey2': -1,
        'prey5': -36,
        'prey6': -36
    }

    assert done == {
        'predator0': False,
        'predator1': False,
        'prey1': False,
        'prey2': False,
        'prey5': True,
        'prey6': True,
        '__all__': False}

    with pytest.raises(AssertionError):
        obs, reward, done, info = sim.step({
            'predator0': {'attack': 1, 'move': np.array([0, 0])},
            'predator1': {'attack': 1, 'move': np.array([0, 0])},
            'prey1': np.array([0, -1]),
            'prey2': np.array([1, 1]),
            'prey5': np.array([1, 1]),
            'prey6': np.array([0, 0]),
        })

    obs, reward, done, info = sim.step({
        'predator0': {'attack': 1, 'move': np.array([0, 0])},
        'predator1': {'attack': 1, 'move': np.array([0, 0])},
        'prey1': np.array([-1, 0]),
        'prey2': np.array([-1, 0]),
    })

    np.testing.assert_array_equal(obs['predator0']['predator1'], np.array([-1, -2,  2]))
    np.testing.assert_array_equal(obs['predator0']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey1'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey2'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey4'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey5'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator0']['prey6'], np.array([0, 0,  0]))

    np.testing.assert_array_equal(obs['predator1']['predator0'], np.array([1, 2,  2]))
    np.testing.assert_array_equal(obs['predator1']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey1'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey2'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey4'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey5'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['predator1']['prey6'], np.array([0, 0,  0]))

    np.testing.assert_array_equal(obs['prey1']['predator0'], np.array([0, 2,  2]))
    np.testing.assert_array_equal(obs['prey1']['predator1'], np.array([-1, 0,  2]))
    np.testing.assert_array_equal(obs['prey1']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey1']['prey2'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(obs['prey1']['prey3'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(obs['prey1']['prey4'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(obs['prey1']['prey5'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(obs['prey1']['prey6'], np.array([0, 0, 0]))

    np.testing.assert_array_equal(obs['prey2']['predator0'], np.array([-1, -1,  2]))
    np.testing.assert_array_equal(obs['prey2']['predator1'], np.array([-2, -3,  2]))
    np.testing.assert_array_equal(obs['prey2']['prey0'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey2']['prey1'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(obs['prey2']['prey3'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey2']['prey4'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey2']['prey5'], np.array([0, 0,  0]))
    np.testing.assert_array_equal(obs['prey2']['prey6'], np.array([0, 0,  0]))

    assert reward == {
        'predator0': 36,
        'predator1': 36,
        'prey1': -36,
        'prey2': -36,
    }

    assert done == {
        'predator0': False,
        'predator1': False,
        'prey1': True,
        'prey2': True,
        '__all__': True}


def test_turn_based_predator_prey_grid():
    np.random.seed(24)
    predators = [Predator(id=f'predator{i}', attack=1, view=0) for i in range(2)]
    prey = [Prey(id=f'prey{i}', view=0) for i in range(7)]
    agents = predators + prey
    sim_config = {
        'region': 6,
        'observation_mode': PredatorPreySimulation.ObservationMode.GRID,
        'agents': agents,
    }
    sim = PredatorPreySimulation.build(sim_config)
    sim = AllStepManager(sim)

    # Little hackish here because I have to explicitly set their values
    obs = sim.reset()
    sim.agents['predator0'].position = np.array([2, 3])
    sim.agents['predator1'].position = np.array([0, 1])
    sim.agents['prey0'].position = np.array([1, 1])
    sim.agents['prey1'].position = np.array([4, 3])
    sim.agents['prey2'].position = np.array([4, 3])
    sim.agents['prey3'].position = np.array([2, 3])
    sim.agents['prey4'].position = np.array([3, 3])
    sim.agents['prey5'].position = np.array([3, 1])
    sim.agents['prey6'].position = np.array([2, 1])
    obs = {agent_id: sim.sim.get_obs(agent_id) for agent_id in sim.agents}

    assert 'predator0' in obs
    assert 'predator0' in obs
    assert 'prey0' in obs
    assert 'prey1' in obs
    assert 'prey2' in obs
    assert 'prey3' in obs
    assert 'prey4' in obs
    assert 'prey5' in obs
    assert 'prey6' in obs

    obs, reward, done, info = sim.step({
        'predator0': {'attack': 1, 'move': np.array([0, 0])},
        'predator1': {'attack': 1, 'move': np.array([0, 0])},
        'prey0': {'move': np.array([1, 1]), 'harvest': 0},
        'prey1': {'move': np.array([0, -1]), 'harvest': 0},
        'prey2': {'move': np.array([1, 1]), 'harvest': 0},
        'prey3': {'move': np.array([0, 0]), 'harvest': 0},
        'prey4': {'move': np.array([-1, 1]), 'harvest': 0},
        'prey5': {'move': np.array([1, 1]), 'harvest': 0},
        'prey6': {'move': np.array([0, 0]), 'harvest': 0},
    })

    assert 'predator0' in obs
    assert 'predator0' in obs
    assert 'prey0' in obs
    assert 'prey1' in obs
    assert 'prey2' in obs
    assert 'prey3' in obs
    assert 'prey4' in obs
    assert 'prey5' in obs
    assert 'prey6' in obs

    assert reward == {
        'predator0': 36,
        'predator1': 36,
        'prey0': -36,
        'prey1': -1,
        'prey2': -1,
        'prey3': -36,
        'prey4': -1,
        'prey5': -1,
        'prey6': 0,
    }

    assert done == {
        'predator0': False,
        'predator1': False,
        'prey0': True,
        'prey1': False,
        'prey2': False,
        'prey3': True,
        'prey4': False,
        'prey5': False,
        'prey6': False,
        '__all__': False}


    with pytest.raises(AssertionError):
        obs, reward, done, info = sim.step({
            'predator0': {'attack': 1, 'move': np.array([0, 0])},
            'predator1': {'attack': 1, 'move': np.array([0, 0])},
            'prey0': {'move': np.array([0, -1]), 'harvest': 0},
            'prey1': {'move': np.array([0, -1]), 'harvest': 0},
            'prey2': {'move': np.array([1, 1]), 'harvest': 0},
            'prey3': {'move': np.array([0, -1]), 'harvest': 0},
            'prey4': {'move': np.array([0, -1]), 'harvest': 0},
            'prey5': {'move': np.array([1, 1]), 'harvest': 0},
            'prey6': {'move': np.array([0, 0]), 'harvest': 0},
        })

    obs, reward, done, info = sim.step({
        'predator0': {'attack': 1, 'move': np.array([0, 0])},
        'predator1': {'attack': 0, 'move': np.array([1, 0])},
        'prey1': {'move': np.array([-1, -1]), 'harvest': 0},
        'prey2': {'move': np.array([-1, 0]), 'harvest': 0},
        'prey4': {'move': np.array([0, -1]), 'harvest': 0},
        'prey5': {'move': np.array([-1, 0]), 'harvest': 0},
        'prey6': {'move': np.array([0, -1]), 'harvest': 0},
    })

    assert 'predator0' in obs
    assert 'predator0' in obs
    assert 'prey1' in obs
    assert 'prey2' in obs
    assert 'prey4' in obs
    assert 'prey5' in obs
    assert 'prey6' in obs

    assert reward == {
        'predator0': 36,
        'predator1': -1,
        'prey1': -1,
        'prey2': -1,
        'prey4': -36,
        'prey5': -1,
        'prey6': -1,
    }

    assert done == {
        'predator0': False,
        'predator1': False,
        'prey1': False,
        'prey2': False,
        'prey4': True,
        'prey5': False,
        'prey6': False,
        '__all__': False}


    with pytest.raises(AssertionError):
        obs, reward, done, info = sim.step({
            'predator0': {'attack': 1, 'move': np.array([0, 0])},
            'predator1': {'attack': 1, 'move': np.array([0, 0])},
            'prey1': {'move': np.array([0, -1]), 'harvest': 0},
            'prey2': {'move': np.array([1, 1]), 'harvest': 0},
            'prey4': {'move': np.array([0, -1]), 'harvest': 0},
            'prey5': {'move': np.array([1, 1]), 'harvest': 0},
            'prey6': {'move': np.array([0, 0]), 'harvest': 0},
        })

    obs, reward, done, info = sim.step({
        'predator0': {'attack': 1, 'move': np.array([0, 0])},
        'predator1': {'attack': 1, 'move': np.array([0, 0])},
        'prey1': {'move': np.array([-1, 0]), 'harvest': 0},
        'prey2': {'move': np.array([-1, 0]), 'harvest': 0},
        'prey5': {'move': np.array([-1, 0]), 'harvest': 0},
        'prey6': {'move': np.array([1, -1]), 'harvest': 0},
    })

    assert 'predator0' in obs
    assert 'predator0' in obs
    assert 'prey1' in obs
    assert 'prey2' in obs
    assert 'prey5' in obs
    assert 'prey6' in obs

    assert reward == {
        'predator0': 36,
        'predator1': 36,
        'prey1': -1,
        'prey2': -1,
        'prey5': -36,
        'prey6': -36,
    }

    assert done == {
        'predator0': False,
        'predator1': False,
        'prey1': False,
        'prey2': False,
        'prey5': True,
        'prey6': True,
        '__all__': False}


    with pytest.raises(AssertionError):
        obs, reward, done, info = sim.step({
            'predator0': {'attack': 1, 'move': np.array([0, 0])},
            'predator1': {'attack': 1, 'move': np.array([0, 0])},
            'prey1': {'move': np.array([0, -1]), 'harvest': 0},
            'prey2': {'move': np.array([1, 1]), 'harvest': 0},
            'prey5': {'move': np.array([1, 1]), 'harvest': 0},
            'prey6': {'move': np.array([0, 0]), 'harvest': 0},
        })

    obs, reward, done, info = sim.step({
        'predator0': {'attack': 1, 'move': np.array([0, 0])},
        'predator1': {'attack': 1, 'move': np.array([0, 0])},
        'prey1': {'move': np.array([-1, 0]), 'harvest': 0},
        'prey2': {'move': np.array([-1, 0]), 'harvest': 0},
    })

    assert 'predator0' in obs
    assert 'predator0' in obs
    assert 'prey1' in obs
    assert 'prey2' in obs

    assert reward == {
        'predator0': 36,
        'predator1': 36,
        'prey1': -36,
        'prey2': -36,
    }

    assert done == {
        'predator0': False,
        'predator1': False,
        'prey1': True,
        'prey2': True,
        '__all__': True}



from gym.spaces import Dict, Discrete, Box
import numpy as np
import pytest

from abmarl.sim.predator_prey import PredatorPreySimulation, PredatorPreySimDistanceObs, \
    PredatorPreySimGridObs
from abmarl.sim.predator_prey import Predator, Prey
from abmarl.sim.modules import GridResources


def test_build_fails():
    with pytest.raises(TypeError): # Abstract class error
        PredatorPreySimulation()
    with pytest.raises(TypeError): # Missing argument
        PredatorPreySimDistanceObs()
    with pytest.raises(TypeError): # Missing argument
        PredatorPreySimGridObs()
    with pytest.raises(TypeError): # Abstract class error
        PredatorPreySimulation({
            'region': None,
            'max_steps': None,
            'agents': None,
            'reward_map': None,
        })


def test_class_attributes():
    assert PredatorPreySimulation.ActionStatus.BAD_MOVE == 0
    assert PredatorPreySimulation.ActionStatus.GOOD_MOVE == 1
    assert PredatorPreySimulation.ActionStatus.NO_MOVE == 2
    assert PredatorPreySimulation.ActionStatus.BAD_ATTACK == 3
    assert PredatorPreySimulation.ActionStatus.GOOD_ATTACK == 4
    assert PredatorPreySimulation.ActionStatus.EATEN == 5

    assert PredatorPreySimulation.ObservationMode.GRID == 0
    assert PredatorPreySimulation.ObservationMode.DISTANCE == 1


def test_builder():
    sim = PredatorPreySimulation.build()
    assert isinstance(sim, PredatorPreySimGridObs)
    assert sim.region == 10
    assert sim.max_steps == 200
    assert sim.reward_map == {
        'predator': {
            PredatorPreySimulation.ActionStatus.BAD_MOVE: -10,
            PredatorPreySimulation.ActionStatus.GOOD_MOVE: -1,
            PredatorPreySimulation.ActionStatus.NO_MOVE: 0,
            PredatorPreySimulation.ActionStatus.BAD_ATTACK: -10,
            PredatorPreySimulation.ActionStatus.GOOD_ATTACK: 100,
        },
        'prey': {
            PredatorPreySimulation.ActionStatus.BAD_MOVE: -10,
            PredatorPreySimulation.ActionStatus.GOOD_MOVE: -1,
            PredatorPreySimulation.ActionStatus.NO_MOVE: 0,
            PredatorPreySimulation.ActionStatus.EATEN: -100,
            PredatorPreySimulation.ActionStatus.BAD_HARVEST: -10,
            PredatorPreySimulation.ActionStatus.GOOD_HARVEST: 10,
        }
    }
    grid_resources = GridResources.build({'region': sim.region})
    assert sim.resources.region == grid_resources.region
    assert sim.resources.coverage == grid_resources.coverage
    assert sim.resources.min_value == grid_resources.min_value
    assert sim.resources.max_value == grid_resources.max_value
    assert sim.resources.revive_rate == grid_resources.revive_rate

    agents = sim.agents
    assert type(agents) == dict
    assert len(agents) == 2
    assert agents['prey0'].id == 'prey0'
    assert type(agents['prey0']) == Prey
    assert agents['prey0'].view == 9
    assert agents['prey0'].harvest_amount == 0.1
    assert agents['prey0'].observation_space == Dict({
        'agents': Box(low=-1, high=2, shape=(19,19), dtype=np.int),
        'resources': Box(-1, sim.resources.max_value, (19,19), np.float),
    })
    assert agents['prey0'].action_space == Dict({
        'move': Box(low=-1.5, high=1.5, shape=(2,)),
        'harvest': Discrete(2)
    })
    assert agents['predator0'].id == 'predator0'
    assert type(agents['predator0']) == Predator
    assert agents['predator0'].view == 9
    assert agents['predator0'].attack == 0
    assert agents['predator0'].observation_space == Dict({
        'agents': Box(low=-1, high=2, shape=(19,19), dtype=np.int),
        'resources': Box(-1, sim.resources.max_value, (19,19), np.float),
    })
    assert agents['predator0'].action_space == Dict({
        'move': Box(low=-1.5, high=1.5, shape=(2,)),
        'attack': Discrete(2)
    })


def test_builder_region():
    sim = PredatorPreySimulation.build({'region': 20})
    assert sim.region == 20
    assert sim.resources.region == 20
    with pytest.raises(TypeError):
        PredatorPreySimulation.build({'region': '12'})
    with pytest.raises(TypeError):
        PredatorPreySimulation.build({'region': -2})

    agents = sim.agents
    assert len(agents) == 2
    assert agents['prey0'].id == 'prey0'
    assert type(agents['prey0']) == Prey
    assert agents['prey0'].view == 19
    assert agents['prey0'].observation_space == Dict({
        'agents': Box(low=-1, high=2, shape=(39,39), dtype=np.int),
        'resources': Box(-1, sim.resources.max_value, (39,39), np.float),
    })
    assert agents['predator0'].id == 'predator0'
    assert type(agents['predator0']) == Predator
    assert agents['predator0'].view == 19
    assert agents['predator0'].attack == 0
    assert agents['predator0'].observation_space == Dict({
        'agents': Box(low=-1, high=2, shape=(39,39), dtype=np.int),
        'resources': Box(-1, sim.resources.max_value, (39,39), np.float),
    })


def test_build_max_steps():
    sim = PredatorPreySimulation.build({'max_steps': 100})
    assert sim.max_steps == 100
    with pytest.raises(TypeError):
        PredatorPreySimulation.build({'max_steps': 12.5})
    with pytest.raises(TypeError):
        PredatorPreySimulation.build({'max_steps': -8})


def test_builder_observation_mode():
    sim = PredatorPreySimulation.build(
        {'observation_mode': PredatorPreySimulation.ObservationMode.DISTANCE}
    )
    assert isinstance(sim, PredatorPreySimDistanceObs)

    agents = sim.agents
    assert type(agents) == dict
    assert len(agents) == 2
    assert agents['prey0'].id == 'prey0'
    assert type(agents['prey0']) == Prey
    assert agents['prey0'].view == 9
    assert agents['prey0'].observation_space == Dict({
        'predator0': Box(-9, 9, (3,), np.int)
    })
    assert agents['predator0'].id == 'predator0'
    assert type(agents['predator0']) == Predator
    assert agents['predator0'].view == 9
    assert agents['predator0'].attack == 0
    assert agents['predator0'].observation_space == Dict({
        'prey0': Box(low=-9, high=9, shape=(3,), dtype=np.int)
    })


def test_builder_rewards():
    rewards = {
        'predator': {
            PredatorPreySimulation.ActionStatus.BAD_MOVE: -2,
            PredatorPreySimulation.ActionStatus.GOOD_MOVE: -1,
            PredatorPreySimulation.ActionStatus.NO_MOVE: 0,
            PredatorPreySimulation.ActionStatus.BAD_ATTACK: -5,
            PredatorPreySimulation.ActionStatus.GOOD_ATTACK: 10,
        },
        'prey': {
            PredatorPreySimulation.ActionStatus.BAD_MOVE: -2,
            PredatorPreySimulation.ActionStatus.GOOD_MOVE: 2,
            PredatorPreySimulation.ActionStatus.NO_MOVE: 1,
            PredatorPreySimulation.ActionStatus.EATEN: -10,
        }
    }
    sim = PredatorPreySimulation.build({'rewards': rewards})
    assert sim.reward_map == rewards
    with pytest.raises(TypeError):
        PredatorPreySimulation.build({'rewards': 12})


def test_builder_resources():
    resources = {
        'coverage': 0.5,
        'min_value': 0.3,
        'max_value': 1.2,
        'revive_rate': 0.1,
    }
    sim = PredatorPreySimulation.build({'resources': resources})
    assert sim.resources.region == sim.region
    assert sim.resources.coverage == resources['coverage']
    assert sim.resources.min_value == resources['min_value']
    assert sim.resources.max_value == resources['max_value']
    assert sim.resources.revive_rate == resources['revive_rate']


def test_builder_agents():
    np.random.seed(24)
    # Create some good agents
    agents = [
        Prey(id='prey0', view=7, move=2),
        Predator(id='predator1', view=3, attack=2),
        Prey(id='prey2', view=5, move=3),
        Predator(id='predator3', view=2, move=2, attack=1),
        Predator(id='predator4', view=0, attack=3)
    ]
    sim = PredatorPreySimulation.build({'agents': agents})

    agents = sim.agents
    for agent in agents.values():
        assert agent.configured

    assert agents['prey0'].observation_space == Dict({
        'agents': Box(-1, 2, (15,15), np.int),
        'resources': Box(-1, sim.resources.max_value, (15,15), np.float),
    })
    assert agents['predator1'].observation_space == Dict({
        'agents': Box(-1, 2, (7, 7), np.int),
        'resources': Box(-1, sim.resources.max_value, (7,7), np.float),
    })
    assert agents['prey2'].observation_space == Dict({
        'agents': Box(-1, 2, (11, 11), np.int),
        'resources': Box(-1, sim.resources.max_value, (11,11), np.float),
    })
    assert agents['predator3'].observation_space == Dict({
        'agents': Box(-1, 2, (5, 5), np.int),
        'resources': Box(-1, sim.resources.max_value, (5,5), np.float),
    })
    assert agents['predator4'].observation_space == Dict({
        'agents': Box(-1, 2, (1, 1), np.int),
        'resources': Box(-1, sim.resources.max_value, (1,1), np.float),
    })

    assert agents['prey0'].action_space == Dict({
        'move': Box(-2.5, 2.5, (2,)),
        'harvest': Discrete(2),
    })
    assert agents['predator1'].action_space == Dict({
        'attack': Discrete(2),
        'move': Box(-1.5, 1.5, (2,)),
    })
    assert agents['prey2'].action_space == Dict({
        'move': Box(-3.5, 3.5, (2,)),
        'harvest': Discrete(2),
    })
    assert agents['predator3'].action_space == Dict({
        'attack': Discrete(2),
        'move': Box(-2.5, 2.5, (2,)),
    })
    assert agents['predator4'].action_space == Dict({
        'attack': Discrete(2),
        'move': Box(-1.5, 1.5, (2,)),
    })


def test_reset_grid_obs():
    np.random.seed(24)
    agents = [
        Prey(id='prey0', view=2),
        Predator(id='predator1', view=4),
        Prey(id='prey2', view=2),
        Predator(id='predator3', view=4),
        Predator(id='predator4', view=4),
    ]
    sim = PredatorPreySimulation.build({'agents': agents})
    sim.reset()

    # Explicitly place the agents
    sim.agents['predator1'].position = np.array([4,4])
    sim.agents['predator3'].position = np.array([3,3])
    sim.agents['predator4'].position = np.array([7,9])
    sim.agents['prey0'].position = np.array([1,1])
    sim.agents['prey2'].position = np.array([3,2])

    assert sim.step_count == 0
    np.testing.assert_array_equal(sim.get_obs('predator1')['agents'], np.array([
        [0., 0., 0., 0., 0., 0., 0., 0., 0.],
        [0., 1., 0., 0., 0., 0., 0., 0., 0.],
        [0., 0., 0., 0., 0., 0., 0., 0., 0.],
        [0., 0., 1., 2., 0., 0., 0., 0., 0.],
        [0., 0., 0., 0., 0., 0., 0., 0., 0.],
        [0., 0., 0., 0., 0., 0., 0., 0., 0.],
        [0., 0., 0., 0., 0., 0., 0., 0., 0.],
        [0., 0., 0., 0., 0., 0., 0., 0., 0.],
        [0., 0., 0., 0., 0., 0., 0., 0., 0.]]
    ))
    np.testing.assert_array_equal(sim.get_obs('predator3')['agents'], np.array([
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1.,  0.,  1.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1.,  0.,  0.,  1.,  0.,  0.,  0.,  0.,  0.],
        [-1.,  0.,  0.,  0.,  0.,  2.,  0.,  0.,  0.],
        [-1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.]]
    ))
    np.testing.assert_array_equal(sim.get_obs('predator4')['agents'], np.array(
        [[ 0.,  0.,  0.,  0.,  0., -1., -1., -1., -1.],
        [ 0.,  0.,  0.,  0.,  0., -1., -1., -1., -1.],
        [ 0.,  0.,  0.,  0.,  0., -1., -1., -1., -1.],
        [ 0.,  0.,  0.,  0.,  0., -1., -1., -1., -1.],
        [ 0.,  0.,  0.,  0.,  0., -1., -1., -1., -1.],
        [ 0.,  0.,  0.,  0.,  0., -1., -1., -1., -1.],
        [ 0.,  0.,  0.,  0.,  0., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.]]
    ))
    np.testing.assert_array_equal(sim.get_obs('prey0')['agents'], np.array(
        [[-1., -1., -1., -1., -1.],
        [-1.,  0.,  0.,  0.,  0.],
        [-1.,  0.,  0.,  0.,  0.],
        [-1.,  0.,  0.,  0.,  0.],
        [-1.,  0.,  0.,  1.,  2.]]
    ))
    np.testing.assert_array_equal(sim.get_obs('prey2')['agents'], np.array([
        [0., 1., 0., 0., 0.],
        [0., 0., 0., 0., 0.],
        [0., 0., 0., 2., 0.],
        [0., 0., 0., 0., 2.],
        [0., 0., 0., 0., 0.]]
    ))


def test_reset_distance_obs():
    np.random.seed(24)
    agents = [
        Prey(id='prey0', view=2),
        Predator(id='predator1', view=4),
        Prey(id='prey2', view=2),
        Predator(id='predator3', view=4),
        Predator(id='predator4', view=4),
    ]
    sim = PredatorPreySimulation.build(
        {'agents': agents, 'observation_mode': PredatorPreySimulation.ObservationMode.DISTANCE}
    )
    sim.reset()

    # Explicitly place the agents
    sim.agents['predator1'].position = np.array([4,4])
    sim.agents['predator3'].position = np.array([3,3])
    sim.agents['predator4'].position = np.array([7,9])
    sim.agents['prey0'].position = np.array([1,1])
    sim.agents['prey2'].position = np.array([3,2])

    assert sim.step_count == 0

    np.testing.assert_array_equal(sim.get_obs('predator1')['predator3'], np.array([-1, -1, 2]))
    np.testing.assert_array_equal(sim.get_obs('predator1')['predator4'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(sim.get_obs('predator1')['prey0'], np.array([-3, -3, 1]))
    np.testing.assert_array_equal(sim.get_obs('predator1')['prey2'], np.array([-1, -2, 1]))

    np.testing.assert_array_equal(sim.get_obs('predator3')['predator1'], np.array([1, 1, 2]))
    np.testing.assert_array_equal(sim.get_obs('predator3')['predator4'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(sim.get_obs('predator3')['prey0'], np.array([-2, -2, 1]))
    np.testing.assert_array_equal(sim.get_obs('predator3')['prey2'], np.array([0, -1, 1]))

    np.testing.assert_array_equal(sim.get_obs('predator4')['predator1'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(sim.get_obs('predator4')['predator3'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(sim.get_obs('predator4')['prey0'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(sim.get_obs('predator4')['prey2'], np.array([0, 0, 0]))

    np.testing.assert_array_equal(sim.get_obs('prey0')['predator1'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(sim.get_obs('prey0')['predator3'], np.array([2, 2, 2]))
    np.testing.assert_array_equal(sim.get_obs('prey0')['predator4'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(sim.get_obs('prey0')['prey2'], np.array([2, 1, 1]))

    np.testing.assert_array_equal(sim.get_obs('prey2')['predator1'], np.array([1, 2, 2]))
    np.testing.assert_array_equal(sim.get_obs('prey2')['predator3'], np.array([0, 1, 2]))
    np.testing.assert_array_equal(sim.get_obs('prey2')['predator4'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(sim.get_obs('prey2')['prey0'], np.array([-2, -1, 1]))


def test_step_grid_obs():
    np.random.seed(24)
    agents = [
        Predator(id='predator0', view=2, attack=1),
        Prey(id='prey1', view=4),
        Prey(id='prey2', view=5)
    ]
    sim = PredatorPreySimulation.build({'agents': agents})
    sim.reset()
    sim.agents['predator0'].position = np.array([2, 3])
    sim.agents['prey1'].position = np.array([0, 7])
    sim.agents['prey2'].position = np.array([1, 1])
    np.testing.assert_array_equal(sim.get_obs('predator0')['agents'], np.array([
        [0., 0., 0., 0., 0.],
        [1., 0., 0., 0., 0.],
        [0., 0., 0., 0., 0.],
        [0., 0., 0., 0., 0.],
        [0., 0., 0., 0., 0.]
    ]))
    np.testing.assert_array_equal(sim.get_obs('prey1')['agents'], np.array([
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0., -1., -1.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0., -1., -1.],
        [ 2.,  0.,  0.,  0.,  0.,  0.,  0., -1., -1.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0., -1., -1.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0., -1., -1.]
    ]))
    np.testing.assert_array_equal(sim.get_obs('prey2')['agents'], np.array([
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1., -1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1., -1., -1.,  0.,  0.,  0.,  2.,  0.,  0.,  0.],
        [-1., -1., -1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1., -1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1., -1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1., -1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.]
    ]))

    action = {
        'predator0': {'move': np.zeros(2), 'attack': 1},
        'prey1': {'move': np.array([0, -1]), 'harvest': 0},
        'prey2': {'move': np.array([-1, 0]), 'harvest': 0},
    }
    sim.step(action)
    np.testing.assert_array_equal(sim.get_obs('predator0')['agents'], np.array([
        [1., 0., 0., 0., 0.],
        [0., 0., 0., 0., 0.],
        [0., 0., 0., 0., 0.],
        [0., 0., 0., 0., 0.],
        [0., 0., 0., 0., 0.]
    ]))
    assert sim.get_reward('predator0') == -10
    assert not sim.get_done('predator0')
    np.testing.assert_array_equal(sim.get_obs('prey1')['agents'], np.array([
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0., -1.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0., -1.],
        [ 0.,  2.,  0.,  0.,  0.,  0.,  0.,  0., -1.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0., -1.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0., -1.]
    ]))
    assert sim.get_reward('prey1') == -1
    assert not sim.get_done('prey1')
    np.testing.assert_array_equal(sim.get_obs('prey2')['agents'], np.array([
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  1.],
        [-1., -1., -1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1., -1., -1.,  0.,  0.,  0.,  2.,  0.,  0.,  0.],
        [-1., -1., -1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1., -1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1., -1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.]
    ]))
    assert sim.get_reward('prey2') == -1
    assert not sim.get_done('prey2')
    assert not sim.get_all_done()

    action = {
        'predator0': {'move': np.array([-1, 0]), 'attack': 0},
        'prey1': {'move': np.array([1, 0]), 'harvest': 0},
        'prey2': {'move': np.array([-1, 0]), 'harvest': 0},
    }
    sim.step(action)
    np.testing.assert_array_equal(sim.get_obs('predator0')['agents'], np.array([
        [-1., -1., -1., -1., -1.],
        [ 1.,  0.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  0.,  0.]]))
    assert sim.get_reward('predator0') == -1
    assert not sim.get_done('predator0')
    np.testing.assert_array_equal(sim.get_obs('prey1')['agents'], np.array([
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0., -1.],
        [ 0.,  2.,  0.,  0.,  0.,  0.,  0.,  0., -1.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0., -1.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0., -1.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0., -1.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0., -1.]]))
    assert sim.get_reward('prey1') == -1
    assert not sim.get_done('prey1')
    np.testing.assert_array_equal(sim.get_obs('prey2')['agents'], np.array([
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1., -1., -1.,  0.,  0.,  0.,  2.,  0.,  0.,  1.],
        [-1., -1., -1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1., -1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1., -1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1., -1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.]]))
    assert sim.get_reward('prey2') == -10
    assert not sim.get_done('prey2')
    assert not sim.get_all_done()

    action = {
        'predator0': {'move': np.array([0,0]), 'attack': 0},
        'prey1': {'move': np.array([0, -1]), 'harvest': 0},
        'prey2': {'move': np.array([0, 1]), 'harvest': 0},
    }
    sim.step(action)
    np.testing.assert_array_equal(sim.get_obs('predator0')['agents'], np.array([
        [-1., -1., -1., -1., -1.],
        [ 0.,  1.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  0.,  1.],
        [ 0.,  0.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  0.,  0.]]))
    assert sim.get_reward('predator0') == 0
    assert not sim.get_done('predator0')
    np.testing.assert_array_equal(sim.get_obs('prey1')['agents'], np.array([
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [ 0.,  1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [ 0.,  0.,  2.,  0.,  0.,  0.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.]]))
    assert sim.get_reward('prey1') == -1
    assert not sim.get_done('prey1')
    np.testing.assert_array_equal(sim.get_obs('prey2')['agents'], np.array([
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1., -1.,  0.,  0.,  0.,  2.,  0.,  1.,  0.,  0.],
        [-1., -1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.]]))
    assert sim.get_reward('prey2') == -1
    assert not sim.get_done('prey2')
    assert not sim.get_all_done()

    action = {
        'predator0': {'move': np.array([0, 1]), 'attack': 0},
        'prey1': {'move': np.array([1, 0]), 'harvest': 0},
        'prey2': {'move': np.array([0, 1]), 'harvest': 0},
    }
    sim.step(action)
    np.testing.assert_array_equal(sim.get_obs('predator0')['agents'], np.array([
        [-1., -1., -1., -1., -1.],
        [ 0.,  1.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  1.,  0.],
        [ 0.,  0.,  0.,  0.,  0.]]))
    assert sim.get_reward('predator0') == -1
    assert not sim.get_done('predator0')
    np.testing.assert_array_equal(sim.get_obs('prey1')['agents'], np.array([
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [ 0.,  0.,  1.,  0.,  0.,  0.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  2.,  0.,  0.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.]]))
    assert sim.get_reward('prey1') == -1
    assert not sim.get_done('prey1')
    np.testing.assert_array_equal(sim.get_obs('prey2')['agents'], np.array([
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1.,  0.,  0.,  0.,  0.,  2.,  0.,  0.,  0.,  0.],
        [-1., -1.,  0.,  0.,  0.,  0.,  0.,  1.,  0.,  0.,  0.],
        [-1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.]]))
    assert sim.get_reward('prey2') == -1
    assert not sim.get_done('prey2')
    assert not sim.get_all_done()

    action = {
        'predator0': {'move': np.zeros(2), 'attack': 1},
        'prey1': {'move': np.array([1, 0]), 'harvest': 0},
        'prey2': {'move': np.array([1, 0]), 'harvest': 0},
    }
    sim.step(action)
    np.testing.assert_array_equal(sim.get_obs('predator0')['agents'], np.array([
        [-1., -1., -1., -1., -1.],
        [ 0.,  0.,  0.,  0.,  0.],
        [ 0.,  1.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  0.,  0.]]))
    assert sim.get_reward('predator0') == 100
    assert not sim.get_done('predator0')
    np.testing.assert_array_equal(sim.get_obs('prey1')['agents'], np.array([
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [ 0.,  0.,  1.,  2.,  0.,  0.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.]]))
    assert sim.get_reward('prey1') == -100
    assert sim.get_done('prey1')
    np.testing.assert_array_equal(sim.get_obs('prey2')['agents'], np.array([
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1.,  0.,  0.,  0.,  0.,  2.,  0.,  0.,  0.,  0.],
        [-1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.]]))
    assert sim.get_reward('prey2') == -1
    assert not sim.get_done('prey2')
    assert not sim.get_all_done()

    action = {
        'predator0': {'move': np.zeros(2), 'attack': 1},
        'prey2': {'move': np.array([0, 1]), 'harvest': 0},
    }
    sim.step(action)
    np.testing.assert_array_equal(sim.get_obs('predator0')['agents'], np.array([
        [-1., -1., -1., -1., -1.],
        [ 0.,  0.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  0.,  0.]]))
    assert sim.get_reward('predator0') == 100
    assert not sim.get_done('predator0')
    np.testing.assert_array_equal(sim.get_obs('prey2')['agents'], np.array([
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1.,  0.,  0.,  0.,  0.,  2.,  0.,  0.,  0.,  0.],
        [-1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.]]))
    assert sim.get_reward('prey2') == -100
    assert sim.get_done('prey2')
    assert sim.get_all_done()


def test_step_distance_obs():
    np.random.seed(24)
    agents = [
        Predator(id='predator0', view=2, attack=1),
        Prey(id='prey1', view=4),
        Prey(id='prey2', view=5)
    ]
    sim = PredatorPreySimulation.build(
        {'agents': agents, 'observation_mode': PredatorPreySimulation.ObservationMode.DISTANCE}
    )
    sim.reset()
    sim.agents['predator0'].position = np.array([2, 3])
    sim.agents['prey1'].position = np.array([0, 7])
    sim.agents['prey2'].position = np.array([1, 1])

    np.testing.assert_array_equal(sim.get_obs('predator0')['prey1'], np.array([ 0, 0, 0]))
    np.testing.assert_array_equal(sim.get_obs('predator0')['prey2'], np.array([-1, -2, 1]))

    np.testing.assert_array_equal(sim.get_obs('prey1')['predator0'], np.array([ 2, -4, 2]))
    np.testing.assert_array_equal(sim.get_obs('prey1')['prey2'], np.array([0, 0, 0]))

    np.testing.assert_array_equal(sim.get_obs('prey2')['predator0'], np.array([ 1, 2, 2]))
    np.testing.assert_array_equal(sim.get_obs('prey2')['prey1'], np.array([0, 0, 0]))


    action = {
        'predator0': {'move': np.zeros(2), 'attack': 1},
        'prey1': np.array([-1, 0]),
        'prey2': np.array([0, 1]),
    }
    sim.step(action)

    np.testing.assert_array_equal(sim.get_obs('predator0')['prey1'], np.array([ 0, 0, 0]))
    np.testing.assert_array_equal(sim.get_obs('predator0')['prey2'], np.array([-1, -1, 1]))
    assert sim.get_reward('predator0') == -10
    assert not sim.get_done('predator0')

    np.testing.assert_array_equal(sim.get_obs('prey1')['predator0'], np.array([ 2, -4, 2]))
    np.testing.assert_array_equal(sim.get_obs('prey1')['prey2'], np.array([0, 0, 0]))
    assert sim.get_reward('prey1') == -10
    assert not sim.get_done('prey1')

    np.testing.assert_array_equal(sim.get_obs('prey2')['predator0'], np.array([ 1, 1, 2]))
    np.testing.assert_array_equal(sim.get_obs('prey2')['prey1'], np.array([-1, 5, 1]))
    assert sim.get_reward('prey2') == -1
    assert not sim.get_done('prey2')
    assert not sim.get_all_done()


    action = {
        'predator0': {'move': np.array([-1, 0]), 'attack': 0},
        'prey1': np.array([0, -1]),
        'prey2': np.array([0, 1]),
    }
    sim.step(action)

    np.testing.assert_array_equal(sim.get_obs('predator0')['prey1'], np.array([ 0, 0, 0]))
    np.testing.assert_array_equal(sim.get_obs('predator0')['prey2'], np.array([0, 0, 1]))
    assert sim.get_reward('predator0') == -1
    assert not sim.get_done('predator0')

    np.testing.assert_array_equal(sim.get_obs('prey1')['predator0'], np.array([ 1, -3, 2]))
    np.testing.assert_array_equal(sim.get_obs('prey1')['prey2'], np.array([1, -3, 1]))
    assert sim.get_reward('prey1') == -1
    assert not sim.get_done('prey1')

    np.testing.assert_array_equal(sim.get_obs('prey2')['predator0'], np.array([ 0, 0, 2]))
    np.testing.assert_array_equal(sim.get_obs('prey2')['prey1'], np.array([-1, 3, 1]))
    assert sim.get_reward('prey2') == -1
    assert not sim.get_done('prey2')
    assert not sim.get_all_done()


    action = {
        'predator0': {'move': np.array([0,0]), 'attack': 0},
        'prey1': np.array([0, -1]),
        'prey2': np.array([0, 1]),
    }
    sim.step(action)

    np.testing.assert_array_equal(sim.get_obs('predator0')['prey1'], np.array([-1, 2, 1]))
    np.testing.assert_array_equal(sim.get_obs('predator0')['prey2'], np.array([0, 1, 1]))
    assert sim.get_reward('predator0') == 0
    assert not sim.get_done('predator0')

    np.testing.assert_array_equal(sim.get_obs('prey1')['predator0'], np.array([ 1, -2, 2]))
    np.testing.assert_array_equal(sim.get_obs('prey1')['prey2'], np.array([1, -1, 1]))
    assert sim.get_reward('prey1') == -1
    assert not sim.get_done('prey1')

    np.testing.assert_array_equal(sim.get_obs('prey2')['predator0'], np.array([ 0, -1, 2]))
    np.testing.assert_array_equal(sim.get_obs('prey2')['prey1'], np.array([-1, 1, 1]))
    assert sim.get_reward('prey2') == -1
    assert not sim.get_done('prey2')
    assert not sim.get_all_done()


    action = {
        'predator0': {'move': np.array([0, 1]), 'attack': 0},
        'prey1': np.array([0, -1]),
        'prey2': np.array([-1, 0]),
    }
    sim.step(action)

    np.testing.assert_array_equal(sim.get_obs('predator0')['prey1'], np.array([-1, 0, 1]))
    np.testing.assert_array_equal(sim.get_obs('predator0')['prey2'], np.array([-1, 0, 1]))
    assert sim.get_reward('predator0') == -1
    assert not sim.get_done('predator0')

    np.testing.assert_array_equal(sim.get_obs('prey1')['predator0'], np.array([ 1, 0, 2]))
    np.testing.assert_array_equal(sim.get_obs('prey1')['prey2'], np.array([0, 0, 1]))
    assert sim.get_reward('prey1') == -1
    assert not sim.get_done('prey1')

    np.testing.assert_array_equal(sim.get_obs('prey2')['predator0'], np.array([1, 0, 2]))
    np.testing.assert_array_equal(sim.get_obs('prey2')['prey1'], np.array([0,0,1]))
    assert sim.get_reward('prey2') == -1
    assert not sim.get_done('prey2')
    assert not sim.get_all_done()


    action = {
        'predator0': {'move': np.zeros(2), 'attack': 1},
        'prey1': np.array([0, 1]),
        'prey2': np.array([0, -1]),
    }
    sim.step(action)

    np.testing.assert_array_equal(sim.get_obs('predator0')['prey1'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(sim.get_obs('predator0')['prey2'], np.array([-1, -1, 1]))
    assert sim.get_reward('predator0') == 100
    assert not sim.get_done('predator0')

    np.testing.assert_array_equal(sim.get_obs('prey1')['predator0'], np.array([ 1, 0, 2]))
    np.testing.assert_array_equal(sim.get_obs('prey1')['prey2'], np.array([0, -1, 1]))
    assert sim.get_reward('prey1') == -100
    assert sim.get_done('prey1')

    np.testing.assert_array_equal(sim.get_obs('prey2')['predator0'], np.array([1, 1, 2]))
    np.testing.assert_array_equal(sim.get_obs('prey2')['prey1'], np.array([0,0,0]))
    assert sim.get_reward('prey2') == -1
    assert not sim.get_done('prey2')
    assert not sim.get_all_done()


    action = {
        'predator0': {'move': np.zeros(2), 'attack': 1},
        'prey2': np.array([1, 0]),
    }
    sim.step(action)

    np.testing.assert_array_equal(sim.get_obs('predator0')['prey1'], np.array([0, 0, 0]))
    np.testing.assert_array_equal(sim.get_obs('predator0')['prey2'], np.array([0, 0, 0]))
    assert sim.get_reward('predator0') == 100
    assert not sim.get_done('predator0')

    np.testing.assert_array_equal(sim.get_obs('prey2')['predator0'], np.array([1, 1, 2]))
    np.testing.assert_array_equal(sim.get_obs('prey2')['prey1'], np.array([0,0,0]))
    assert sim.get_reward('prey2') == -100
    assert sim.get_done('prey2')
    assert sim.get_all_done()


def test_attack_distances():
    region = 5
    predators = [Predator(id='predator0')]
    # predators = [{'id': 'predator0', 'view': region-1, 'move': 1, 'attack': 0}]
    prey = [Prey(id='prey{}'.format(i)) for i in range(3)]
    # prey = [{'id': 'prey' + str(i), 'view': region-1, 'move': 1} for i in range(3)]
    agents = predators + prey
    config = {'region': region, 'agents': agents}
    sim = PredatorPreySimulation.build(config)
    sim.reset()
    sim.agents['predator0'].position = np.array([2, 2])
    sim.agents['prey0'].position = np.array([2, 2])
    sim.agents['prey1'].position = np.array([1, 1])
    sim.agents['prey2'].position = np.array([0, 0])
    assert sim.agents['predator0'].attack == 0
    action_dict = {
        'predator0': {'move': np.zeros([0, 0]), 'attack': 1},
        'prey0': {'move': np.zeros(2), 'harvest': 0},
        'prey1': {'move': np.zeros(2), 'harvest': 0},
        'prey2': {'move': np.zeros(2), 'harvest': 0},
    }


    sim.step(action_dict)
    assert sim.get_reward('predator0') == \
        sim.reward_map['predator'][PredatorPreySimulation.ActionStatus.GOOD_ATTACK]
    assert sim.get_reward('prey0') == sim.reward_map['prey'][
        PredatorPreySimulation.ActionStatus.EATEN
    ]
    assert sim.get_reward('prey1') == sim.reward_map['prey'][
        PredatorPreySimulation.ActionStatus.NO_MOVE
    ]
    assert sim.get_reward('prey2') == sim.reward_map['prey'][
        PredatorPreySimulation.ActionStatus.NO_MOVE
    ]
    assert not sim.get_done('predator0')
    assert sim.get_done('prey0')
    assert not sim.get_done('prey1')
    assert not sim.get_done('prey2')
    assert not sim.get_all_done()

    del action_dict['prey0']
    sim.step(action_dict)
    assert sim.get_reward('predator0') == \
        sim.reward_map['predator'][PredatorPreySimulation.ActionStatus.BAD_ATTACK]
    assert sim.get_reward('prey1') == sim.reward_map['prey'][
        PredatorPreySimulation.ActionStatus.NO_MOVE
    ]
    assert sim.get_reward('prey2') == sim.reward_map['prey'][
        PredatorPreySimulation.ActionStatus.NO_MOVE
    ]
    assert not sim.get_done('predator0')
    assert not sim.get_done('prey1')
    assert not sim.get_done('prey2')
    assert not sim.get_all_done()


    sim.agents['predator0'].attack = 1

    sim.step(action_dict)
    assert sim.get_reward('predator0') == \
        sim.reward_map['predator'][PredatorPreySimulation.ActionStatus.GOOD_ATTACK]
    assert sim.get_reward('prey1') == sim.reward_map['prey'][
        PredatorPreySimulation.ActionStatus.EATEN
    ]
    assert sim.get_reward('prey2') == sim.reward_map['prey'][
        PredatorPreySimulation.ActionStatus.NO_MOVE
    ]
    assert not sim.get_done('predator0')
    assert sim.get_done('prey1')
    assert not sim.get_done('prey2')
    assert not sim.get_all_done()

    del action_dict['prey1']
    sim.step(action_dict)
    assert sim.get_reward('predator0') == \
        sim.reward_map['predator'][PredatorPreySimulation.ActionStatus.BAD_ATTACK]
    assert sim.get_reward('prey2') == sim.reward_map['prey'][
        PredatorPreySimulation.ActionStatus.NO_MOVE
    ]
    assert not sim.get_done('predator0')
    assert not sim.get_done('prey2')
    assert not sim.get_all_done()


    sim.agents['predator0'].attack = 2
    sim.step(action_dict)
    assert sim.get_reward('predator0') == \
        sim.reward_map['predator'][PredatorPreySimulation.ActionStatus.GOOD_ATTACK]
    assert sim.get_reward('prey2') == \
        sim.reward_map['prey'][PredatorPreySimulation.ActionStatus.EATEN]
    assert not sim.get_done('predator0')
    assert sim.get_done('prey2')
    assert sim.get_all_done()


def test_diagonal_moves():
    np.random.seed(24)
    region = 3
    prey = [Prey(id='prey{}'.format(i)) for i in range(12)]
    sim = PredatorPreySimulation.build({'agents': prey, 'region': region})
    sim.reset()
    for prey in sim.agents.values():
        prey.position = np.array([1,1])


    action = {
        'prey0': {'move': np.array([0, 1]), 'harvest': 0},
        'prey1': {'move': np.array([0, 1]), 'harvest': 0},
        'prey2': {'move': np.array([0, -1]), 'harvest': 0},
        'prey3': {'move': np.array([0, -1]), 'harvest': 0},
        'prey4': {'move': np.array([-1, 0]), 'harvest': 0},
        'prey5': {'move': np.array([-1, 0]), 'harvest': 0},
        'prey6': {'move': np.array([1, 0]), 'harvest': 0},
        'prey7': {'move': np.array([1, 0]), 'harvest': 0},
        'prey8': {'move': np.array([1, 1]), 'harvest': 0},
        'prey9': {'move': np.array([1, -1]), 'harvest': 0},
        'prey10': {'move': np.array([-1, -1]), 'harvest': 0},
        'prey11': {'move': np.array([-1, 1]), 'harvest': 0},
    }
    sim.step(action)
    assert sim.get_reward('prey0') == \
        sim.reward_map['prey'][PredatorPreySimulation.ActionStatus.GOOD_MOVE]
    assert sim.get_reward('prey1') == \
        sim.reward_map['prey'][PredatorPreySimulation.ActionStatus.GOOD_MOVE]
    assert sim.get_reward('prey2') == \
        sim.reward_map['prey'][PredatorPreySimulation.ActionStatus.GOOD_MOVE]
    assert sim.get_reward('prey3') == \
        sim.reward_map['prey'][PredatorPreySimulation.ActionStatus.GOOD_MOVE]
    assert sim.get_reward('prey4') == \
        sim.reward_map['prey'][PredatorPreySimulation.ActionStatus.GOOD_MOVE]
    assert sim.get_reward('prey5') == \
        sim.reward_map['prey'][PredatorPreySimulation.ActionStatus.GOOD_MOVE]
    assert sim.get_reward('prey6') == \
        sim.reward_map['prey'][PredatorPreySimulation.ActionStatus.GOOD_MOVE]
    assert sim.get_reward('prey7') == \
        sim.reward_map['prey'][PredatorPreySimulation.ActionStatus.GOOD_MOVE]
    assert sim.get_reward('prey8') == \
        sim.reward_map['prey'][PredatorPreySimulation.ActionStatus.GOOD_MOVE]
    assert sim.get_reward('prey9') == \
        sim.reward_map['prey'][PredatorPreySimulation.ActionStatus.GOOD_MOVE]
    assert sim.get_reward('prey10') == \
        sim.reward_map['prey'][PredatorPreySimulation.ActionStatus.GOOD_MOVE]
    assert sim.get_reward('prey11') == \
        sim.reward_map['prey'][PredatorPreySimulation.ActionStatus.GOOD_MOVE]

    np.testing.assert_array_equal([agent.position for agent in sim.agents.values()], [
        np.array([1, 2]),
        np.array([1, 2]),
        np.array([1, 0]),
        np.array([1, 0]),
        np.array([0, 1]),
        np.array([0, 1]),
        np.array([2, 1]),
        np.array([2, 1]),
        np.array([2, 2]),
        np.array([2, 0]),
        np.array([0, 0]),
        np.array([0, 2]),
    ])


    action = {
        'prey0': {'move': np.array([1, 1]), 'harvest': 0},
        'prey1': {'move': np.array([-1, 1]), 'harvest': 0},
        'prey2': {'move': np.array([1, -1]), 'harvest': 0},
        'prey3': {'move': np.array([-1, -1]), 'harvest': 0},
        'prey4': {'move': np.array([-1, 1]), 'harvest': 0},
        'prey5': {'move': np.array([-1, -1]), 'harvest': 0},
        'prey6': {'move': np.array([1, 1]), 'harvest': 0},
        'prey7': {'move': np.array([1, -1]), 'harvest': 0},
        'prey8': {'move': np.array([1, 1]), 'harvest': 0},
        'prey9': {'move': np.array([1, -1]), 'harvest': 0},
        'prey10': {'move': np.array([-1, -1]), 'harvest': 0},
        'prey11': {'move': np.array([-1, 1]), 'harvest': 0},
    }
    sim.step(action)
    assert sim.get_reward('prey0') == sim.reward_map['prey'][
        PredatorPreySimulation.ActionStatus.BAD_MOVE
    ]
    assert sim.get_reward('prey1') == sim.reward_map['prey'][
        PredatorPreySimulation.ActionStatus.BAD_MOVE
    ]
    assert sim.get_reward('prey2') == sim.reward_map['prey'][
        PredatorPreySimulation.ActionStatus.BAD_MOVE
    ]
    assert sim.get_reward('prey3') == sim.reward_map['prey'][
        PredatorPreySimulation.ActionStatus.BAD_MOVE
    ]
    assert sim.get_reward('prey4') == sim.reward_map['prey'][
        PredatorPreySimulation.ActionStatus.BAD_MOVE
    ]
    assert sim.get_reward('prey5') == sim.reward_map['prey'][
        PredatorPreySimulation.ActionStatus.BAD_MOVE
    ]
    assert sim.get_reward('prey6') == sim.reward_map['prey'][
        PredatorPreySimulation.ActionStatus.BAD_MOVE
    ]
    assert sim.get_reward('prey7') == sim.reward_map['prey'][
        PredatorPreySimulation.ActionStatus.BAD_MOVE
    ]
    assert sim.get_reward('prey8') == sim.reward_map['prey'][
        PredatorPreySimulation.ActionStatus.BAD_MOVE
    ]
    assert sim.get_reward('prey9') == sim.reward_map['prey'][
        PredatorPreySimulation.ActionStatus.BAD_MOVE
    ]
    assert sim.get_reward('prey10') == \
        sim.reward_map['prey'][PredatorPreySimulation.ActionStatus.BAD_MOVE]
    assert sim.get_reward('prey11') == \
        sim.reward_map['prey'][PredatorPreySimulation.ActionStatus.BAD_MOVE]

    np.testing.assert_array_equal([agent.position for agent in sim.agents.values()], [
        np.array([1, 2]),
        np.array([1, 2]),
        np.array([1, 0]),
        np.array([1, 0]),
        np.array([0, 1]),
        np.array([0, 1]),
        np.array([2, 1]),
        np.array([2, 1]),
        np.array([2, 2]),
        np.array([2, 0]),
        np.array([0, 0]),
        np.array([0, 2]),
    ])


def test_multi_move():
    np.random.seed(24)
    region = 8
    prey = [Prey(id='prey{}'.format(i), move=4) for i in range(4)]
    sim = PredatorPreySimulation.build({'agents': prey, 'region': region})
    sim.reset()
    sim.agents['prey0'].position = np.array([2, 3])
    sim.agents['prey1'].position = np.array([0, 7])
    sim.agents['prey2'].position = np.array([1, 1])
    sim.agents['prey3'].position = np.array([1, 4])

    action = {agent_id: agent.action_space.sample() for agent_id, agent in sim.agents.items()}
    action = {
        'prey0': {'move': np.array([-2, 3]), 'harvest': 0},
        'prey1': {'move': np.array([4, 0]), 'harvest': 0},
        'prey2': {'move': np.array([2, 1]), 'harvest': 0},
        'prey3': {'move': np.array([3, -2]), 'harvest': 0},
    }
    sim.step(action)

    np.testing.assert_array_equal(sim.agents['prey0'].position, [0,6])
    np.testing.assert_array_equal(sim.agents['prey1'].position, [4,7])
    np.testing.assert_array_equal(sim.agents['prey2'].position, [3,2])
    np.testing.assert_array_equal(sim.agents['prey3'].position, [4,2])


def test_done_on_max_steps():
    agents = [Prey(id=f'prey{i}') for i in range(2)]
    sim = PredatorPreySimulation.build({'max_steps': 4, 'agents': agents})
    sim.reset()
    for i in range(4):
        sim.step({agent_id: agent.action_space.sample() for agent_id, agent in sim.agents.items()})
    assert sim.get_all_done()


def test_with_resources():
    np.random.seed(24)
    agents = [
        Prey(id='prey0', view=2, harvest_amount=0.3),
        Prey(id='prey1'),
        Predator(id='predator0', view=1)
    ]
    sim = PredatorPreySimulation.build({'region': 5, 'agents': agents})
    sim.reset()

    np.allclose(sim.get_obs('predator0')['resources'], np.array([
        [0.        , 0.19804811, 0.        ],
        [0.16341112, 0.58086431, 0.4482749 ],
        [0.        , 0.38637824, 0.78831386]
    ]))
    np.allclose(sim.get_obs('prey0')['resources'], np.array([
        [ 0.19804811, 0.        ,  0.42549817,  0.9438245 , -1.        ],
        [ 0.58086431,  0.4482749 ,  0.40239527,  0.31349653, -1.        ],
        [ 0.38637824,  0.78831386,  0.33666274,  0.71590738, -1.        ],
        [ 0.6264872 ,  0.65159097,  0.84080142,  0.24749604, -1.        ],
        [ 0.86455522,  0.        ,  0.        ,  0.        , -1.        ]]))
    np.allclose(sim.get_obs('prey1')['resources'], np.array([
        [-1., -1., -1., -1.        , -1.        , -1.        , -1.        , -1.        , -1.],
        [-1., -1., -1., -1.        , -1.        , -1.        , -1.        , -1.        , -1.],
        [-1., -1., -1., -1.        , -1.        , -1.        , -1.        , -1.        , -1.],
        [-1., -1., -1., -1.        , -1.        , -1.        , -1.        , -1.        , -1.],
        [-1., -1., -1.,  0.        ,  0.19804811,  0.        ,  0.42549817,  0.9438245 , -1.],
        [-1., -1., -1.,  0.16341112,  0.58086431,  0.4482749 ,  0.40239527,  0.31349653, -1.],
        [-1., -1., -1.,  0.        ,  0.38637824,  0.78831386,  0.33666274,  0.71590738, -1.],
        [-1., -1., -1.,  0.        ,  0.6264872 ,  0.65159097,  0.84080142,  0.24749604, -1.],
        [-1., -1., -1.,  0.67319672,  0.86455522,  0.        ,  0.        ,  0.        , -1.]
    ]))
    sim.step({
        'prey0': {'move': np.zeros, 'harvest': 1},
        'prey1': {'move': np.zeros, 'harvest': 1},
    })

    np.allclose(sim.get_obs('predator0')['resources'], np.array([
        [0.        , 0.19804811, 0.        ],
        [0.16341112, 0.58086431, 0.4482749 ],
        [0.        , 0.38637824, 0.78831386]
    ]))
    np.allclose(sim.get_obs('prey0')['resources'], np.array([
        [ 0.19804811,  0.        ,  0.42549817,  0.9438245 , -1.        ],
        [ 0.58086431,  0.4482749 ,  0.40239527,  0.31349653, -1.        ],
        [ 0.38637824,  0.78831386,  0.07666274,  0.71590738, -1.        ],
        [ 0.6264872 ,  0.65159097,  0.84080142,  0.24749604, -1.        ],
        [ 0.86455522,  0.        ,  0.        ,  0.        , -1.        ]]))
    np.allclose(sim.get_obs('prey1')['resources'], np.array([
        [-1., -1., -1., -1.        , -1.        , -1.        , -1.        , -1.        , -1.],
        [-1., -1., -1., -1.        , -1.        , -1.        , -1.        , -1.        , -1.],
        [-1., -1., -1., -1.        , -1.        , -1.        , -1.        , -1.        , -1.],
        [-1., -1., -1., -1.        , -1.        , -1.        , -1.        , -1.        , -1.],
        [-1., -1., -1.,  0.        ,  0.19804811, 0.         ,  0.42549817 , 0.9438245 , -1.],
        [-1., -1., -1.,  0.16341112,  0.58086431, 0.4482749  ,  0.40239527 , 0.31349653, -1.],
        [-1., -1., -1.,  0.        ,  0.38637824, 0.78831386 ,  0.07666274 , 0.71590738, -1.],
        [-1., -1., -1.,  0.        ,  0.6264872 , 0.65159097 ,  0.84080142 , 0.24749604, -1.],
        [-1., -1., -1.,  0.67319672,  0.86455522, 0.         ,  0.         , 0.        , -1.]
    ]))
    assert sim.get_reward('prey0') == sim.reward_map['prey'][sim.ActionStatus.GOOD_HARVEST]
    assert sim.get_reward('prey1') == sim.reward_map['prey'][sim.ActionStatus.BAD_HARVEST]



import numpy as np


class GridResources:
    """
    GridResources provides resources that exist on the grid and can be consumed
    by agents in the simulation via their "harvest" action. The resources will
    replenish over time. The grid is covered up to some coverage percentage, and
    the initial value of the resources on each cell are random between the minimum
    and maximum values.

        max_value: double
            The maximum value that a resource can reach. Default 1.0
        min_value: double
            The minimum value that a resource can reach and still be able
            to regenerate itself over time. If the resource value falls below the
            min_value, then the resource will not revive itself. Default 0.1
        revive_rate: double
            The rate of revival for each of the resources. Default 0.04
        coverage: double
            The ratio of the map that is covered with a resource. Default 0.75.
    """
    def __init__(self, config):
        self.region = config['region']
        self.coverage = config['coverage']
        self.min_value = config['min_value']
        self.max_value = config['max_value']
        self.revive_rate = config['revive_rate']

    def reset(self, **kwargs):
        """
        Reset the grid and cover with resources.
        """
        coverage_filter = np.zeros((self.region, self.region))
        coverage_filter[np.random.uniform(0, 1, (self.region, self.region)) < self.coverage] = 1.
        self.resources = np.multiply(
            np.random.uniform(self.min_value, self.max_value, (self.region, self.region)),
            coverage_filter
        )

    def harvest(self, location, amount, **kwargs):
        """
        Process harvesting a certain amount at a certain location. Return the amount
        that was actually harvested here.
        """
        # Process all the harvesting
        if self.resources[location] - amount >= 0.:
            actual_amount_harvested = amount
        else:
            actual_amount_harvested = self.resources[location]
        self.resources[location] = max([0., self.resources[location] - amount])

        return actual_amount_harvested

    def regrow(self, **kwargs):
        """
        Process the regrowth, which is done according to the revival rate.
        """
        self.resources[self.resources >= self.min_value] += self.revive_rate
        self.resources[self.resources >= self.max_value] = self.max_value

    def render(self, *args, fig=None, **kwargs):
        draw_now = fig is None
        if draw_now:
            from matplotlib import pyplot as plt
            fig = plt.gcf()
        import seaborn as sns

        fig.clear()
        ax = fig.gca()
        ax = sns.heatmap(np.flipud(self.resources), ax=ax, cmap='Greens')

        if draw_now:
            plt.plot()
            plt.pause(1e-17)

        return ax

    @classmethod
    def build(cls, sim_config={}):
        config = {
            'region': 10,
            'max_value': 1.,
            'min_value': 0.1,
            'revive_rate': 0.04,
            'coverage': 0.75
        }
        for key, value in config.items():
            config[key] = sim_config.get(key, value)
        return cls(config)
